from ..exceptions import ParseError
from typing import Dict, List, Any, Tuple
from itertools import chain
import warnings

from .base import (
    ElementBase,
    NamedPoseBearing,
    Pose,
    vector3,
    PoseBearing,
    StringElement,
    FloatElement,
)
from .model import Model
from .physics import Physics
from .light import Light
from .frame import Frame
from .scene import Scene
from .actor import Actor
from .joint import Joint
from .plugin import Plugin
from .state import State
from .include import Include
from .population import Population
from .atmosphere import Atmosphere
from .gui import Gui
from .... import transform as tf


class World(ElementBase):
    """A simulation world

    The world element is a container for all information related to a single
    simulation.

    Parameters
    ----------
    name : str
        .. versionadded:: SDFormat v1.2

        The name of the world. It must be unique among all worlds in the same
        :class:`Sdf` container.
    audio: World.Audio
        .. versionadded:: SDFormat v1.4

        The simulations global audio properties.
    wind : World.Wind
        .. versionadded:: SDFormat v1.6

        The direction and strength of the wind. Currently only constant wind
        is supported.

    includes : "World.Include"
        .. versionadded:: SDFormat v1.4

        References to other SDF files that contain fragments
        (:class:`Model`s, :class:`Light`s, :class:`Actor`s) to include in the
        simulation.

    gravity : str
        .. versionadded:: SDFormat v1.6

        The gravity in m/s^2, expressed in a coordinate frame defined by the
        spherical_coordinates attribute. The default is "0 0 -9.8".

    magnetic_field : str
        .. versionadded:: SDFormat v1.6

        The magnetic field vector in Tesla, expressed in a coordinate frame
        defined by the spherical_coordinates attribute. The default is
        "5.5645e-6 22.8758e-6 -42.3884e-6".
    atmosphere : World.Atmosphere
        .. versionadded:: SDFormat v1.6

        The atmosphere model to use.
    gui : World.Gui
        Layout of the Ignition Gazebo GUI. (Applications other than Gazebo will
        likely not need this.)
    physics_engine : Physics
        .. deprecated:: SDFormat v1.6
            Use `physics_engines` instead.

        Configuration parameters of the dynamics engine.
    physics_engines : List[Physics]
        .. versionadded:: SDFormat v1.6

        Configuration parameters of the dynamics engine.
    scene : Scene
        Ambience configuration of the world (skybox, ambient light, etc.)
    lights : List[Light]
        A list of light sources of the world.
    frames : List[Frame]
        .. versionadded:: SDFormat v1.7

        A list of reference frames used to ease the positioning of objects.
    models : List[Model]
        A list of models used in this simulation.
    actors : List[Actor]
        A list of actors used in this simulation.
    plugins : List[Plugin]
        A list of plugins used to customize the runtime behavior of the
        simulation.
    joints : List[Joint]
        .. deprecated:: SDFormat v1.7
            To attach models to the world, use the canonical_link kwarg
            instead. All other usage of `joints` has no replacement.

        A list of connections between two :class:`Link`s that constrains
        their relative movement.
    road : List[World.Road]
        A list of roads used in this simulation.
    spherical_coordinates : World.SphericalCoordinates
        .. versionadded:: SDFormat v1.4

        Simulator root frame location.

    state : List[State]
        The current state of this simulation.
    population : List["Population"]
        .. versionadded:: SDFormat v1.5

        A group of models that is procedually inserted into the simulation.
    sdf_version : str
        The SDFormat version to use when constructing this element.

    Attributes
    ----------
    name : str
        The name of the world.
    audio: World.Audio
        The simulations global audio properties.
    wind : World.Wind
        The direction and strength of the wind.
    gravity : np.ndarray
        The gravity in m/s^2, expressed in a coordinate frame defined by the
        spherical_coordinates attribute.
    magnetic_field : np.ndarray
        The magnetic field vector in Tesla, expressed in a coordinate frame
        defined by the spherical_coordinates attribute.
    atmosphere : World.Atmosphere
        The atmosphere model to use.
    gui : World.Gui
        Layout of the Ignition Gazebo GUI.
    physics_engines : List[Physics]
        Configuration parameters of the dynamics engine(s).
    scene : Scene
        Ambience configuration of the world (skybox, ambient light, etc.)
    lights : List[Light]
        A list of light sources of the world.
    frames : List[Frame]
        A list of reference frames used to ease the positioning of objects.
    models : List[Model]
        A list of models used in this simulation.
    actors : List[Actor]
        A list of actors used in this simulation.
    plugins : List[Plugin]
        A list of plugins used to customize the runtime behavior of the
        simulation.
    road : List[World.Road]
        A list of roads used in this simulation.
    spherical_coordinates : World.SphericalCoordinates
        Simulator root frame location.
    state : List[State]
        The current state of this simulation.

    physics_engine : Physics
        .. deprecated:: SDFormat v1.6
            Use `physics_engines` instead.
    joints : List[Joint]
        .. deprecated:: SDFormat v1.7
            Use `Model.joints` instead and set `Model.joints.parent` to
            `world`.

    Notes
    -----
    The `includes` kwarg is resolved upon construction and the included fragmens
    are appended to `World.models`, `Worlds.actors`, or `Worlds.lights`
    respectively.

    The `populations` kwarg is resolved upon construction and generated models
    are appended to the list of `Worlds.models`.

    """

    def __init__(
        self,
        *,
        name: str = "",
        audio: "World.Audio" = None,
        wind: "World.Wind" = None,
        includes: List[Include],
        gravity: str = "0 0 -9.8",
        magnetic_field: str = "5.5645e-6 22.8758e-6 -42.3884e-6",
        atmosphere: "Atmosphere" = None,
        gui: "Gui" = None,
        physics_engine: Physics = None,
        physics_engines: List[Physics] = None,
        scene: Scene,
        lights: List[Light],
        frames: List[Frame],
        models: List[Model],
        actors: List[Actor],
        plugins: List[Plugin],
        joints: List[Joint] = None,
        roads: List["World.Road"],
        spherical_coordinates: "World.SphericalCoordinates" = None,
        states: List[State],
        populations: List["World.WorldPopulation"],
        sdf_version: str = "1.8",
    ) -> None:
        super().__init__(sdf_version=sdf_version)
        self.name = name
        self.audio = World.Audio(sdf_version=sdf_version) if audio is None else audio
        self.wind = World.Wind(sdf_version=sdf_version) if wind is None else wind
        self.gravity = vector3(gravity)
        self.magnetic_field = vector3(magnetic_field)
        self.atmosphere = (
            Atmosphere(sdf_version=sdf_version) if atmosphere is None else atmosphere
        )
        self.gui = Gui(sdf_version=sdf_version) if gui is None else gui
        if sdf_version in ["1.0", "1.2", "1.3", "1.4", "1.5"]:
            if physics_engines is not None:
                raise ValueError(
                    "`World.physics_engines` requires an SDF version newer than 1.5."
                )
            self.physics_engines = [] if physics_engine is None else [physics_engine]
        else:
            if physics_engine is not None:
                raise ValueError(
                    "`World.physics_engine` is deprecated. Use `physics_engines` instead."
                )
            self.physics_engines = [] if physics_engines is None else physics_engines
        self.scene = scene
        self.lights = lights
        self.frames = frames
        self.models = models
        self.actors = actors
        self.plugins = plugins
        self.populations = populations
        if sdf_version in ["1.0", "1.2", "1.3", "1.4", "1.5", "1.6"]:
            self._joints = [] if joints is None else joints
        elif joints is not None:
            raise ValueError(
                "`World.joints` is deprecated."
                "To connect a model to the world use `Model.canonical_link` instead."
            )
        else:
            self._joints = list()
        self.road = roads
        self.spherical_coordinates = (
            World.SphericalCoordinates()
            if spherical_coordinates is None
            else spherical_coordinates
        )
        self.state = states

        for include in includes:
            fragment = include.resolve()
            if isinstance(fragment, Model):
                self.models.append(fragment)
            elif isinstance(fragment, Actor):
                self.actors.append(fragment)
            else:
                self.lights.append(fragment)

        for frame in self.frames:
            if frame.attached_to is None:
                frame.attached_to = "world"

        el: PoseBearing
        pose_bearing: List[PoseBearing] = [
            # lights,
            self.frames,
            self.models,
            self._joints,
        ]
        for el in chain(*pose_bearing):
            if el.pose.relative_to is None:
                el.pose.relative_to = "world"

    @property
    def joints(self):
        warnings.warn(
            "`World.joints` has been deprecated. Use `Model.joints` with"
            " `Model.joints.parent = 'world'` instead.",
            DeprecationWarning,
        )
        return self._joints

    @classmethod
    def from_specific(cls, specific: Any, *, version: str) -> "World":
        world_args = {
            "includes": [
                Include.from_specific(x, version=version) for x in specific.include
            ],
            "scene": Scene.from_specific(specific.scene, version=version),
        }

        elements_with_default: Dict[str, ElementBase] = {
            "name": StringElement,
            "gravity": StringElement,
            "magnetic_field": StringElement,
            "audio": World.Audio,
            "wind": World.Wind,
            "atmosphere": Atmosphere,
            "gui": Gui,
            "spherical_coordinates": World.SphericalCoordinates,
        }
        list_elements: Dict[str, Tuple[str, ElementBase]] = {
            "include": ("includes", Include),
            "light": ("lights", Light),
            "frame": ("frames", Frame),
            "model": ("models", Model),
            "actor": ("actors", Actor),
            "plugin": ("plugins", Plugin),
            "road": ("roads", World.Road),
            "state": ("states", State),
            "population": ("populations", World.WorldPopulation),
        }

        args = cls._prepare_standard_args(
            specific, elements_with_default, list_elements, version=version
        )
        world_args.update(args)

        if version in ["1.0", "1.1", "1.2", "1.3", "1.4", "1.5"]:
            world_args["physics_engine"] = Physics.from_specific(
                specific.physics, version=version
            )
        else:
            world_args["physics_engines"] = [
                Physics.from_specific(x, version=version) for x in specific.physics
            ]

        if version in ["1.0", "1.1", "1.2", "1.3", "1.4", "1.5", "1.6"]:
            world_args["joints"] = [
                Joint.from_specific(x, version=version) for x in specific.joint
            ]

        return World(**world_args, sdf_version=version)

    def declared_frames(self) -> Dict[str, tf.Frame]:
        world_frame = tf.Frame(3, name=self.name)
        declared_frames = {"world": world_frame}

        for el in self.models:
            model_frames = el.declared_frames()
            declared_frames[el.name] = model_frames["__model__"]
            for name, frame in model_frames.items():
                declared_frames[f"{el.name}::{name}"] = frame

        for el in chain(self.frames, self._joints):
            declared_frames.update(el.declared_frames())

        return declared_frames

    def to_static_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        *,
        seed: int = None,
        shape: Tuple[int] = ...,
        axis: int = -1,
    ) -> tf.Frame:
        for model in self.models:
            scope = {
                name.split("::", 1)[1]: frame
                for name, frame in declared_frames.items()
                if name.startswith(f"{model.name}::")
            }
            scope["world"] = declared_frames["world"]
            model.to_static_graph(scope, seed=seed, shape=shape, axis=axis)

        for model in self.models:
            link = model.pose.to_tf_link()
            parent_name = model.pose.relative_to
            child_name = f"{model.name}::{model.placement_frame}"

            parent = declared_frames[parent_name]
            child = declared_frames[child_name]
            link(child, parent)

        for el in chain(self.frames, self._joints):
            link: tf.Link = el.pose.to_tf_link()
            parent_name = el.pose.relative_to
            child_name = el.name

            parent = declared_frames[parent_name]
            child = declared_frames[child_name]
            link(child, parent)

            el.to_static_graph(declared_frames, seed=seed, shape=shape, axis=axis)

        return declared_frames["world"]

    def to_dynamic_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        *,
        seed: int = None,
        shape: Tuple[int] = (3,),
        axis: int = -1,
        apply_state: bool = True,
        _scaffolding: Dict[str, tf.Frame] = None,
    ) -> tf.Frame:
        if _scaffolding is None:
            _scaffolding = self.declared_frames()
            self.to_static_graph(_scaffolding, seed=seed, shape=shape, axis=axis)

        for model in self.models:
            scaffold_scope = {
                name.split("::", 1)[1]: frame
                for name, frame in _scaffolding.items()
                if name.startswith(f"{model.name}::")
            }
            scaffold_scope["world"] = _scaffolding["world"]

            scope = {
                name.split("::", 1)[1]: frame
                for name, frame in declared_frames.items()
                if name.startswith(f"{model.name}::")
            }
            scope["world"] = declared_frames["world"]
            model.to_dynamic_graph(
                scope,
                seed=seed,
                shape=shape,
                axis=axis,
                apply_state=apply_state,
                _scaffolding=scaffold_scope,
            )

        for el in chain(self.frames, self._joints):
            el.to_dynamic_graph(
                declared_frames,
                seed=seed,
                shape=shape,
                axis=axis,
                apply_state=apply_state,
                _scaffolding=_scaffolding,
            )

        return declared_frames["world"]

    class Audio(ElementBase):
        """Global audio properties.

        Parameters
        ----------
        uri : str
            Device to use for audio playback. If None it will be set to
            "default".

        Attributes
        ----------
        uri : str
            The audio device to use.

        """

        def __init__(self, *, device: str = "default", sdf_version: str) -> None:
            super().__init__(sdf_version=sdf_version)
            self.uri = device

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
            if specific is None:
                return World.Audio(sdf_version=version)
            return World.Audio(device=specific.device, sdf_version=version)

    class Wind(ElementBase):
        """The type and properties of the wind in the world.

        Parameters
        ----------
        linear_velocity : str
            The direction and strength of the wind. Default is "0 0 0", meaning
            no wind.

        Attributes
        ----------
        linear_velocity : np.ndarray
            The direction and strength of the wind's velocity.

        """

        def __init__(self, *, linear_velocity: str = "0 0 0", sdf_version: str) -> None:
            self.linear_velocity = vector3(linear_velocity)

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
            if specific is None:
                return World.Wind(sdf_version=version)
            return World.Wind(
                linear_velocity=specific.linear_velocity, sdf_version=version
            )

    class Road(ElementBase):
        def __init__(self, *, sdf_version: str) -> None:
            warnings.warn("`Road` has not been implemented yet.")
            super().__init__(sdf_version=sdf_version)

    class SphericalCoordinates(ElementBase):
        """Simulator root frame location.

        SphericalCoordnates are used to specify the location of the simulation's
        root frame on Earth using a chosen surface model.

        Parameters
        ----------
        surface_model : str
            Name of planetary surface model, used to determine the surface
            altitude at a given latitude and longitude. The default is an
            ellipsoid model of the earth based on the WGS-84 standard. It is
            used in Gazebo's GPS sensor implementation.
        world_frame_orientation : str
            This field identifies how Gazebo world frame is aligned in
            Geographical sense.  The final Gazebo world frame orientation is
            obtained by rotating a frame aligned with following notation by the
            field heading_deg (Note that heading_deg corresponds to positive yaw
            rotation in the NED frame, so it's inverse specifies positive
            Z-rotation in ENU or NWU). Options are:
                - ENU (East-North-Up; default)
                - NED (North-East-Down)
                - NWU (North-West-Up)
            For example, world frame specified by setting
            world_orientation="ENU" and heading_deg=-90° is effectively
            equivalent to NWU with heading of 0°.
        latitude_deg : float
            Geodetic latitude at origin of gazebo reference frame, specified in
            units of degrees. The default is 0.0.
        longitude_deg : float
            Longitude at origin of gazebo reference frame, specified in units of
            degrees. The default is 0.0.
        elevation : float
            Elevation of origin of gazebo reference frame, specified in meters.
            The default is 0.0.
        heading_deg : float
            Heading offset of gazebo reference frame, measured as angle between
            East and gazebo x axis, or equivalently, the angle between North and
            gazebo y axis. The angle is specified in degrees. The default is
            0.0.

        Attributes
        ----------
        surface_model : str
            The name of planetary surface model.
        world_frame_orientation : str
            Reference orientation for frame alignment.
        latitude_deg : float
            Geodetic latitude at origin of the root frame.
        longitude_deg : float
            Longitude at origin of the root frame.
        elevation : float
            Elevation of origin of the root frame.
        heading_deg : float
            Angle (in degrees) between the root's y-axis and magnetic north.
        """

        def __init__(
            self,
            *,
            surface_model: str = "EARTH_WGS84",
            world_frame_orientation: str = "ENU",
            latitude_deg: float = 0.0,
            longitude_deg: float = 0.0,
            elevation: float = 0.0,
            heading_deg: float = 0.0,
            sdf_version: str,
        ) -> None:
            super().__init__(sdf_version=sdf_version)
            self.surface_model = surface_model
            self.world_frame_orientation = world_frame_orientation
            self.latitude_deg = latitude_deg
            self.longitude_deg = longitude_deg
            self.elevation = elevation
            self.heading_deg = heading_deg

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "World":
            elements_with_default: Dict[str, ElementBase] = {
                "surface_model": StringElement,
                "world_frame_orientation": StringElement,
                "latitude_deg": FloatElement,
                "longitude_deg": FloatElement,
                "elevation": FloatElement,
                "heading_deg": FloatElement,
            }
            generic_args = cls._prepare_standard_args(
                specific, elements_with_default, version=version
            )

            return cls(**generic_args, sdf_version=version)

    class WorldPopulation(Population):
        # def __init__(
        #     self,
        #     *,
        #     name: str,
        #     pose: Pose = None,
        #     model_count: int = 1,
        #     distribution: "Distribution",
        #     box: "Box" = None,
        #     cylinder: "Cylinder" = None,
        #     model: Model,
        #     frames: List[Frame] = None,
        # ) -> None:
        #     super().__init__(name=name, pose=pose)
        #     self.name=name
        #     self.model_count = model_count
        #     self.distribution = distribution
        #     self.box = box
        #     self.cylinder = cylinder
        #     self.model = model
        #     self.frames = frames

        #     if frames is None:
        #         self.frames = list()

        class Distribution(ElementBase):
            def __init__(
                self,
                *,
                kind: str = "random",
                rows: int = 1,
                cols: int = 1,
                step: Tuple[float] = (0.5, 0.5, 0),
            ):
                self.type = kind
                self.rows = rows
                self.cols = cols
                self.step = step

        class Box(ElementBase):
            def __init__(self, *, size: Tuple[float] = (1, 1, 1)) -> None:
                self.size = size

        class Cylinder(ElementBase):
            def __init__(self, *, radius: float = 1, length: float = 1) -> None:
                self.radius = radius
                self.length = length



from dataclasses import dataclass, field
from typing import List, Optional
from .actor import Actor
from .joint import Joint
from .light import Light
from .model import Model
from .physics import Physics
from .scene import Scene
from .state import State

__NAMESPACE__ = "sdformat/v1.2/world.xsd"


@dataclass
class World:
    """
    The world element encapsulates an entire world description including:
    models, scene, physics, joints, and plugins.

    Parameters
    ----------
    gui:
    physics: The physics tag specifies the type and properties of the
        dynamics engine.
    scene: Specifies the look of the environment.
    light: The light element describes a light source.
    model: The model element defines a complete robot or any other
        physical object.
    actor:
    plugin: A plugin is a dynamically loaded chunk of code. It can exist
        as a child of world, model, and sensor.
    joint: A joint connections two links with kinematic and dynamic
        properties.
    road:
    state:
    name: Unique name of the world
    """

    class Meta:
        name = "world"

    gui: Optional["World.Gui"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    physics: Optional[Physics] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    scene: Optional[Scene] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    light: List[Light] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    model: List[Model] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    actor: List[Actor] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    plugin: List["World.Plugin"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    joint: List[Joint] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    road: List["World.Road"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    state: List[State] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )

    @dataclass
    class Gui:
        camera: Optional["World.Gui.Camera"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        fullscreen: bool = field(
            default=False,
            metadata={
                "type": "Attribute",
            },
        )

        @dataclass
        class Camera:
            view_controller: str = field(
                default="orbit",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            pose: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            track_visual: Optional["World.Gui.Camera.TrackVisual"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            name: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )

            @dataclass
            class TrackVisual:
                name: str = field(
                    default="__default__",
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                min_dist: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                max_dist: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

    @dataclass
    class Plugin:
        """A plugin is a dynamically loaded chunk of code.

        It can exist as a child of world, model, and sensor.

        Parameters
        ----------
        any_element: This is a special element that should not be
            specified in an SDFormat file. It automatically copies child
            elements into the SDFormat element so that a plugin can
            access the data.
        name: A unique name for the plugin, scoped to its parent.
        filename: Name of the shared library to load. If the filename is
            not a full path name, the file will be searched for in the
            configuration paths.
        """

        any_element: List[object] = field(
            default_factory=list,
            metadata={
                "type": "Wildcard",
                "namespace": "##any",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )
        filename: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

    @dataclass
    class Road:
        """
        Parameters
        ----------
        width: Width of the road
        point: A series of points define the path of the road.
        name: Name of the road
        """

        width: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        point: List[str] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
                "min_occurs": 1,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )



from dataclasses import dataclass, field
from typing import List, Optional
from .model import Model

__NAMESPACE__ = "sdformat/v1.3/state.xsd"


@dataclass
class State:
    """
    Parameters
    ----------
    sim_time: Simulation time stamp of the state [seconds nanoseconds]
    wall_time: Wall time stamp of the state [seconds nanoseconds]
    real_time: Real time stamp of the state [seconds nanoseconds]
    insertions: A list of new model names
    deletions: A list of deleted model names
    model: Model state
    world_name: Name of the world this state applies to
    """

    class Meta:
        name = "state"

    sim_time: str = field(
        default="0 0",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "white_space": "collapse",
            "pattern": r"\d+ \d+",
        },
    )
    wall_time: str = field(
        default="0 0",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "white_space": "collapse",
            "pattern": r"\d+ \d+",
        },
    )
    real_time: str = field(
        default="0 0",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "white_space": "collapse",
            "pattern": r"\d+ \d+",
        },
    )
    insertions: Optional["State.Insertions"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    deletions: Optional["State.Deletions"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    model: List["State.Model"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    world_name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )

    @dataclass
    class Insertions:
        """
        A list of new model names.

        Parameters
        ----------
        model: The model element defines a complete robot or any other
            physical object.
        """

        model: List[Model] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )

    @dataclass
    class Deletions:
        """
        A list of deleted model names.

        Parameters
        ----------
        name: The name of a deleted model
        """

        name: List[str] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
                "min_occurs": 1,
            },
        )

    @dataclass
    class Model:
        """
        Model state.

        Parameters
        ----------
        pose: Pose of the model
        joint: Joint angle
        link: Link state
        name: Name of the model
        """

        pose: str = field(
            default="0 0 0 0 0 0",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        joint: List["State.Model.Joint"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        link: List["State.Model.Link"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

        @dataclass
        class Joint:
            """
            Joint angle.

            Parameters
            ----------
            angle: Angle of an axis
            name: Name of the joint
            """

            angle: List["State.Model.Joint.Angle"] = field(
                default_factory=list,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "min_occurs": 1,
                },
            )
            name: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )

            @dataclass
            class Angle:
                """
                Parameters
                ----------
                value:
                axis: Index of the axis.
                """

                value: Optional[float] = field(
                    default=None,
                    metadata={
                        "required": True,
                    },
                )
                axis: Optional[int] = field(
                    default=None,
                    metadata={
                        "type": "Attribute",
                        "required": True,
                    },
                )

        @dataclass
        class Link:
            """
            Link state.

            Parameters
            ----------
            pose: Pose of the link relative to the model
            velocity: Velocity of the link
            acceleration: Acceleration of the link
            wrench: Force applied to the link
            collision: Collision state
            name: Name of the link
            """

            pose: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            velocity: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            acceleration: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            wrench: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            collision: List["State.Model.Link.Collision"] = field(
                default_factory=list,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            name: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )

            @dataclass
            class Collision:
                """
                Collision state.

                Parameters
                ----------
                pose: Pose of the link relative to the model
                name: Name of the collision
                """

                pose: str = field(
                    default="0 0 0 0 0 0",
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                        "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                    },
                )
                name: Optional[str] = field(
                    default=None,
                    metadata={
                        "type": "Attribute",
                        "required": True,
                    },
                )



from .. import transform as tf
from numpy.typing import ArrayLike
from typing import List, Callable, Union
import numpy as np
from scipy.optimize import minimize_scalar
from itertools import cycle
from scipy.optimize import OptimizeResult

joint = Union[tf.PrismaticJoint, tf.RotationalJoint]


def ccd(
    pointA: ArrayLike,
    pointB: ArrayLike,
    frameA: tf.Frame,
    frameB: tf.Frame,
    cycle_links: List[joint],
    *,
    metric: Callable[[np.ndarray, np.ndarray], float] = None,
    tol: float = 1e-3,
    maxiter: int = 500,
    line_search_maxiter: int = 500,
) -> List[np.ndarray]:
    """Cyclic Coordinate Descent.

    .. note::
        This function will modify the passed-in frame graph as a side effect.

    .. versionadded:: 0.7.0

    This function adjusts the parameters of the links in ``cycle_links`` such that a
    point that has ``pointA`` as its representation in ``frameA`` has ``pointB``
    as representation in ``frameB``. Parameters are fitted in a cyclical
    fashion, i.e., by repeatedly iterating over ``cycle_links``. Each time the
    respective link's parameter is updated to minimize the distance between
    ``pointB`` and ``pointA``'s representation in ``frameB``. Distance is
    measured in frameB using euclidian distance or a custom metric if provided.

    Parameters
    ----------
    pointA : ArrayLike
        The representation of the point in frameA.
    pointB : ArrayLike
        The desired representation of the point in frameB.
    frameA : tf.Frame
        The frame in which pointA is represented.
    frameB : tf.Frame
        The frame in which pointB is represented.
    cycle_links: List[joint]
        A list of 1DoF joints which should be adjusted to make pointA and pointB
        valid representations of the same point.
    metric : Callable
        A function that takes two points (expressed in frameB) and computs the
        distance between them. Its signature is ``metric(transformed_point,
        pointB) -> distance``. If None, the euclidian distance will be used.
    tol : float
        Absolute tolerance for termination.
    maxiter : int
        The maximum number of cycles to perform.
    line_search_maxiter : int
        The maximum number of iterations to use when optimizing a single joint.

    Returns
    -------
    joint_values : List[float]
        The final parameters of each joint.

    Notes
    -----
    Joint limits (min/max) are enforced as hard constraints.

    The current implementation is a naive python implementation and not very
    optimized. PRs improving performance are welcome :)

    """

    joints = cycle(cycle_links)
    joint_values = [0] * len(cycle_links)

    if metric is None:
        metric = lambda x, y: np.linalg.norm(x - y)

    links = frameA.links_between(frameB)

    def optimizely(x, current_joint):
        current_joint.param = x

        current_point = pointA
        for link in links:
            current_point = link.transform(current_point)

        return metric(current_point, pointB)

    for _ in range(maxiter * len(cycle_links)):
        distance = metric(frameA.transform(pointA, frameB), pointB)

        if distance <= tol:
            break

        current_joint = next(joints)

        result: OptimizeResult = minimize_scalar(
            lambda x: optimizely(x, current_joint),
            bounds=(current_joint.lower_limit, current_joint.upper_limit),
            method="bounded",
            options={"maxiter": line_search_maxiter},
        )

        if not result.success:
            raise RuntimeError(f"IK failed. Reason: {result.message}")

        current_joint.param = result.x
    else:
        raise RuntimeError(f"IK exceeded maxiter.")

    for idx in range(len(cycle_links)):
        joint_values[idx] = cycle_links[idx].param

    return joint_values



import numpy as np
from numpy.typing import ArrayLike

from .base import Frame
from ._utils import vector_project


def scale(vector: ArrayLike, scalar: ArrayLike) -> np.ndarray:
    """Scale each dimension of a vector.

    Multiplies each dimension of ``vector`` with the matching dimension of
    ``scalar``. If necessary, ``scalar`` will be broadcasted.

    Parameters
    ----------
    vector : ArrayLike
        A vector to be scaled.
    scalar : ArrayLike
        A vector representing the amount by which to scale each dimension.

    Returns
    -------
    scaled : ArrayLike
        A vector where each dimension is scaled by scalar.

    Notes
    -----
    Exists for completeness. It may be cleaner to simply write
    ``scalar * vector`` instead.

    """
    vector = np.asarray(vector)
    scalar = np.asarray(scalar)

    return scalar * vector


def translate(vector: ArrayLike, direction: ArrayLike) -> np.ndarray:
    """Translate a vector along direction.

    Parameters
    ----------
    vector : ArrayLike
        The vector to be translated.
    direction : ArrayLike
        A vector describing the translation.

    Returns
    -------
    translated_vector : ArrayLike
        The translated vector.

    Notes
    -----
    Exists for completeness. It may be cleaner to simply write
    ``vector + direction`` instead.

    """

    return vector + direction


def rotate(vector: ArrayLike, u: ArrayLike, v: ArrayLike, *, axis=-1) -> np.ndarray:
    """Rotate a vector in the u,v plane.

    Rotates a vector by reflecting it twice. The plane of rotation
    is given by the u-v-plane and the angle of rotation is two times
    the angle from u to v.

    Parameters
    ----------
    vector : ArrayLike
        The vector to be rotated.
    u : ArrayLike
        The first of the two axes defining the plane of rotation
    v : ArrayLike
        The second of the two axes defining the plane of rotation
    axis : int
        The axis along which to compute the reflection. Default: -1.

    Returns
    -------
    rotated_vector : np.ndarray
        The vector rotated in the u-v-plane by two times the angle
        from u to v.

    Notes
    -----
    The angle of rotation is given by the angle between the two vectors that
    define the plane of rotation. The orientation of the rotation is from u
    towards v, and the amount of rotation is twice the angle.

    The scale of u and/or v does not influence the rotation.

    """

    vector = np.asarray(vector)
    u = np.asarray(u)
    v = np.asarray(v)

    # implemented as rotation by two reflections
    return reflect(reflect(vector, u, axis=axis), v, axis=axis)


def reflect(vector: ArrayLike, direction: ArrayLike, *, axis=-1) -> np.ndarray:
    """Reflect a vector along a line defined by direction.

    Parameters
    ----------
    vector : ArrayLike
        The vector to be reflected.
    direction : ArrayLike
        The vector describing the direction along which the reflection takes place.
    axis : int
        The axis along which to compute the reflection. Default: -1.

    Returns
    -------
    reflected_vector : ArrayLike
        The reflected vector.

    Notes
    -----
    The length of direction does not influence the result of the reflection.

    """

    # from: https://en.wikipedia.org/wiki/Reflection_(mathematics)#Reflection_through_a_hyperplane_in_n_dimensions
    vector = np.asarray(vector)
    direction = np.asarray(direction)

    return vector - 2 * vector_project(vector, direction, axis=axis)


def shear(
    vector: ArrayLike, direction: ArrayLike, amount: ArrayLike, *, axis=-1
) -> np.ndarray:
    """Displaces a vector along direction by the scalar product of vector and amount.

    A shear displaces a vector in a fixed direction by the vector's scalar
    projection onto a second vector (amount) scaled by the length of that second
    vector. If amount and direction are orthogonal, the result is a shear. If
    amount and direction are parallel, the result is a stretch.

    Parameters
    ----------
    vector : ArrayLike
        The vector to be sheared.
    direction : ArrayLike
        The direction along which to apply the shear.
    amount : ArrayLike
        The axis that determines the amount to shear by.
    axis : int
        The axis along with to compute the shear.

    Returns
    -------
    sheared : np.ndarray
        The sheared vector.

    Notes
    -----
    If direction is not normalized the resulting shear factor will be scaled by
    the length (euclidian norm) of direction.

    """

    vector = np.asarray(vector)
    direction = np.asarray(direction)
    amount = np.asarray(amount)

    tmp1 = np.sum(vector * amount, axis=axis)

    return vector + tmp1 * direction


def as_affine_matrix(from_frame: Frame, to_frame: Frame, *, axis: int = -1):
    """Transformation Matrix between two frames at a given point.

    Given two frames ``from_frame`` and ``to_frame`` that represent affine space
    and are connected by a sequence of linear transformations, compute a matrix
    representation of the transformation.

    Parameters
    ----------
    from_frame : tf.Frame
        The parent frame.
    to_frame : tf.Frame
        The child frame.
    axis : int
        The axis along which computation takes place. All other axes are considered
        batch dimensions.

    Returns
    -------
    matrix : np.ndarray
        The matrix representation of the transformation. It will have the shape
        ``(batch_shape, to_frame.ndim, from_frame.ndim)``.

    Notes
    -----
    The matrix representation will only be accurate if the transformation chain
    between the two given frames is linear.

    The

    """

    if axis != -1:
        raise NotImplementedError("Axis is not implemented yet.")

    basis_set = np.eye(from_frame.ndim)
    basis_set[:, -1] = 1

    mapped_basis = from_frame.transform(basis_set, to_frame).T

    # normalize affine matrix
    scaling = mapped_basis[-1, :]
    mapped_basis /= scaling[None, :]
    mapped_basis[..., :-1] -= mapped_basis[..., -1][..., None]

    return mapped_basis



from dataclasses import dataclass, field
from typing import List, Optional
from .actor import Actor
from .joint import Joint
from .light import Light
from .model import Model
from .physics import Physics
from .scene import Scene
from .state import State

__NAMESPACE__ = "sdformat/v1.0/world.xsd"


@dataclass
class World:
    class Meta:
        name = "world"

    gui: Optional["World.Gui"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    physics: Optional[Physics] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    scene: Optional[Scene] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    light: List[Light] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    model: List[Model] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    actor: List[Actor] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    plugin: List["World.Plugin"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    joint: List[Joint] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    road: List["World.Road"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    state: List[State] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )

    @dataclass
    class Gui:
        camera: Optional["World.Gui.Camera"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        fullscreen: bool = field(
            default=False,
            metadata={
                "type": "Attribute",
            },
        )

        @dataclass
        class Camera:
            view_controller: str = field(
                default="oribit",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            origin: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            track_visual: Optional["World.Gui.Camera.TrackVisual"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            name: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )

            @dataclass
            class TrackVisual:
                name: str = field(
                    default="__default__",
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                min_dist: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                max_dist: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

    @dataclass
    class Plugin:
        any_element: List[object] = field(
            default_factory=list,
            metadata={
                "type": "Wildcard",
                "namespace": "##any",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )
        filename: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

    @dataclass
    class Road:
        width: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        point: List[str] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
                "min_occurs": 1,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )



from dataclasses import dataclass, field
from typing import Optional
from .geometry import Geometry

__NAMESPACE__ = "sdformat/v1.8/collision.xsd"


@dataclass
class Collision:
    """The collision properties of a link.

    Note that this can be different from the visual properties of a
    link, for example, simpler collision models are often used to reduce
    computation time.

    Parameters
    ----------
    laser_retro: intensity value returned by laser sensor.
    max_contacts: Maximum number of contacts allowed between two
        entities. This value overrides the max_contacts element defined
        in physics.
    pose: A position(x,y,z) and orientation(roll, pitch yaw) with
        respect to the frame named in the relative_to attribute.
    geometry: The shape of the visual or collision object.
    surface: The surface parameters
    name: Unique name for the collision element within the scope of the
        parent link.
    """

    class Meta:
        name = "collision"

    laser_retro: float = field(
        default=0.0,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    max_contacts: int = field(
        default=10,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    pose: Optional["Collision.Pose"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    geometry: Optional[Geometry] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    surface: Optional["Collision.Surface"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )

    @dataclass
    class Pose:
        """
        Parameters
        ----------
        value:
        relative_to: Name of frame relative to which the pose is
            applied.
        """

        value: str = field(
            default="0 0 0 0 0 0",
            metadata={
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        relative_to: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
            },
        )

    @dataclass
    class Surface:
        """
        The surface parameters.
        """

        bounce: Optional["Collision.Surface.Bounce"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        friction: Optional["Collision.Surface.Friction"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        contact: Optional["Collision.Surface.Contact"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        soft_contact: Optional["Collision.Surface.SoftContact"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )

        @dataclass
        class Bounce:
            """
            Parameters
            ----------
            restitution_coefficient: Bounciness coefficient of
                restitution, from [0...1], where 0=no bounciness.
            threshold: Bounce capture velocity, below which effective
                coefficient of restitution is 0.
            """

            restitution_coefficient: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            threshold: float = field(
                default=100000.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

        @dataclass
        class Friction:
            """
            Parameters
            ----------
            torsional: Parameters for torsional friction
            ode: ODE friction parameters
            bullet:
            """

            torsional: Optional["Collision.Surface.Friction.Torsional"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            ode: Optional["Collision.Surface.Friction.Ode"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            bullet: Optional["Collision.Surface.Friction.Bullet"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )

            @dataclass
            class Torsional:
                """
                Parameters for torsional friction.

                Parameters
                ----------
                coefficient: Torsional friction coefficient, unitless
                    maximum ratio of           tangential stress to
                    normal stress.
                use_patch_radius: If this flag is true,
                    torsional friction is calculated using the
                    "patch_radius" parameter.           If this flag is
                    set to false,           "surface_radius" (R) and
                    contact depth (d)           are used to compute the
                    patch radius as sqrt(R*d).
                patch_radius: Radius of contact patch surface.
                surface_radius: Surface radius on the point of contact.
                ode: Torsional friction parameters for ODE
                """

                coefficient: float = field(
                    default=1.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                use_patch_radius: bool = field(
                    default=True,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                patch_radius: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                surface_radius: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                ode: Optional["Collision.Surface.Friction.Torsional.Ode"] = field(
                    default=None,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                    },
                )

                @dataclass
                class Ode:
                    """
                    Torsional friction parameters for ODE.

                    Parameters
                    ----------
                    slip: Force dependent slip for torsional friction,
                        equivalent to inverse of viscous damping
                        coefficient             with units of
                        rad/s/(Nm).             A slip value of 0 is
                        infinitely viscous.
                    """

                    slip: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )

            @dataclass
            class Ode:
                """
                ODE friction parameters.

                Parameters
                ----------
                mu: Coefficient of friction in first friction pyramid
                    direction,           the unitless maximum ratio of
                    force in first friction pyramid           direction
                    to normal force.
                mu2: Coefficient of friction in second friction pyramid
                    direction,           the unitless maximum ratio of
                    force in second friction pyramid           direction
                    to normal force.
                fdir1: Unit vector specifying first friction pyramid
                    direction in           collision-fixed reference
                    frame.           If the friction pyramid model is in
                    use,           and this value is set to a unit
                    vector for one of the           colliding surfaces,
                    the ODE Collide callback function will align the
                    friction pyramid directions           with a
                    reference frame fixed to that collision surface.
                    If both surfaces have this value set to a vector of
                    zeros,           the friction pyramid directions
                    will be aligned with the world frame.           If
                    this value is set for both surfaces, the behavior is
                    undefined.
                slip1: Force dependent slip in first friction pyramid
                    direction,           equivalent to inverse of
                    viscous damping coefficient           with units of
                    m/s/N.           A slip value of 0 is infinitely
                    viscous.
                slip2: Force dependent slip in second friction pyramid
                    direction,           equivalent to inverse of
                    viscous damping coefficient           with units of
                    m/s/N.           A slip value of 0 is infinitely
                    viscous.
                """

                mu: float = field(
                    default=1.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                mu2: float = field(
                    default=1.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                fdir1: str = field(
                    default="0 0 0",
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                        "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                    },
                )
                slip1: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                slip2: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

            @dataclass
            class Bullet:
                """
                Parameters
                ----------
                friction: Coefficient of friction in first friction
                    pyramid direction,           the unitless maximum
                    ratio of force in first friction pyramid
                    direction to normal force.
                friction2: Coefficient of friction in second friction
                    pyramid direction,           the unitless maximum
                    ratio of force in second friction pyramid
                    direction to normal force.
                fdir1: Unit vector specifying first friction pyramid
                    direction in           collision-fixed reference
                    frame.           If the friction pyramid model is in
                    use,           and this value is set to a unit
                    vector for one of the           colliding surfaces,
                    the friction pyramid directions will be aligned
                    with a reference frame fixed to that collision
                    surface.           If both surfaces have this value
                    set to a vector of zeros,           the friction
                    pyramid directions will be aligned with the world
                    frame.           If this value is set for both
                    surfaces, the behavior is undefined.
                rolling_friction: Coefficient of rolling friction
                """

                friction: float = field(
                    default=1.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                friction2: float = field(
                    default=1.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                fdir1: str = field(
                    default="0 0 0",
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                        "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                    },
                )
                rolling_friction: float = field(
                    default=1.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

        @dataclass
        class Contact:
            """
            Parameters
            ----------
            collide_without_contact: Flag to disable contact force
                generation, while still allowing collision checks and
                contact visualization to occur.
            collide_without_contact_bitmask: Bitmask for collision
                filtering when collide_without_contact is on
            collide_bitmask: Bitmask for collision filtering. This will
                override collide_without_contact. Parsed as 16-bit
                unsigned integer.
            category_bitmask: Bitmask for category of collision
                filtering. Collision happens if ((category1 &amp;
                collision2) | (category2 &amp; collision1)) is not zero.
                If not specified, the category_bitmask should be
                interpreted as being the same as collide_bitmask. Parsed
                as 16-bit unsigned integer.
            poissons_ratio: Poisson's ratio is the unitless ratio
                between transverse and axial strain.         This value
                must lie between (-1, 0.5).  Defaults to 0.3 for typical
                steel.         Note typical silicone elastomers have
                Poisson's ratio near 0.49 ~ 0.50.          For
                reference, approximate values for Material:(Young's
                Modulus, Poisson's Ratio)         for some of the
                typical materials are:           Plastic:  (1e8 ~ 3e9
                Pa,  0.35 ~ 0.41),           Wood:     (4e9 ~ 1e10 Pa,
                0.22 ~ 0.50),           Aluminum: (7e10 Pa,       0.32 ~
                0.35),           Steel:    (2e11 Pa,       0.26 ~ 0.31).
            elastic_modulus: Young's Modulus in SI derived unit Pascal.
                Defaults to -1.  If value is less or equal to zero,
                contact using elastic modulus (with Poisson's Ratio) is
                disabled.          For reference, approximate values for
                Material:(Young's Modulus, Poisson's Ratio)         for
                some of the typical materials are:           Plastic:
                (1e8 ~ 3e9 Pa,  0.35 ~ 0.41),           Wood:     (4e9 ~
                1e10 Pa, 0.22 ~ 0.50),           Aluminum: (7e10 Pa,
                0.32 ~ 0.35),           Steel:    (2e11 Pa,       0.26 ~
                0.31).
            ode: ODE contact parameters
            bullet: Bullet contact parameters
            """

            collide_without_contact: bool = field(
                default=False,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            collide_without_contact_bitmask: int = field(
                default=1,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            collide_bitmask: int = field(
                default=65535,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            category_bitmask: int = field(
                default=65535,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            poissons_ratio: float = field(
                default=0.3,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            elastic_modulus: float = field(
                default=-1.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            ode: Optional["Collision.Surface.Contact.Ode"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            bullet: Optional["Collision.Surface.Contact.Bullet"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )

            @dataclass
            class Ode:
                """
                ODE contact parameters.

                Parameters
                ----------
                soft_cfm: Soft constraint force mixing.
                soft_erp: Soft error reduction parameter
                kp: dynamically "stiffness"-equivalent coefficient for
                    contact joints
                kd: dynamically "damping"-equivalent coefficient for
                    contact joints
                max_vel: maximum contact correction velocity truncation
                    term.
                min_depth: minimum allowable depth before contact
                    correction impulse is applied
                """

                soft_cfm: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                soft_erp: float = field(
                    default=0.2,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                kp: float = field(
                    default=1000000000000.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                kd: float = field(
                    default=1.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                max_vel: float = field(
                    default=0.01,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                min_depth: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

            @dataclass
            class Bullet:
                """
                Bullet contact parameters.

                Parameters
                ----------
                soft_cfm: Soft constraint force mixing.
                soft_erp: Soft error reduction parameter
                kp: dynamically "stiffness"-equivalent coefficient for
                    contact joints
                kd: dynamically "damping"-equivalent coefficient for
                    contact joints
                split_impulse: Similar to ODE's max_vel implementation.
                    See
                    http://bulletphysics.org/mediawiki-1.5.8/index.php/BtContactSolverInfo#Split_Impulse
                    for more information.
                split_impulse_penetration_threshold: Similar to ODE's
                    max_vel implementation.  See
                    http://bulletphysics.org/mediawiki-1.5.8/index.php/BtContactSolverInfo#Split_Impulse
                    for more information.
                """

                soft_cfm: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                soft_erp: float = field(
                    default=0.2,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                kp: float = field(
                    default=1000000000000.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                kd: float = field(
                    default=1.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                split_impulse: bool = field(
                    default=True,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                split_impulse_penetration_threshold: float = field(
                    default=-0.01,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

        @dataclass
        class SoftContact:
            """
            Parameters
            ----------
            dart: soft contact pamameters based on paper:
                http://www.cc.gatech.edu/graphics/projects/Sumit/homepage/papers/sigasia11/jain_softcontacts_siga11.pdf
            """

            dart: Optional["Collision.Surface.SoftContact.Dart"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )

            @dataclass
            class Dart:
                """
                soft contact pamameters based on paper:              http://www
                .cc.gatech.edu/graphics/projects/Sumit/homepage/papers/sigasia1
                1/jain_softcontacts_siga11.pdf.

                Parameters
                ----------
                bone_attachment: This is variable k_v in the soft
                    contacts paper.  Its unit is N/m.
                stiffness: This is variable k_e in the soft contacts
                    paper.  Its unit is N/m.
                damping: Viscous damping of point velocity in body
                    frame.  Its unit is N/m/s.
                flesh_mass_fraction: Fraction of mass to be distributed
                    among deformable nodes.
                """

                bone_attachment: float = field(
                    default=100.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                stiffness: float = field(
                    default=100.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                damping: float = field(
                    default=10.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                flesh_mass_fraction: float = field(
                    default=0.05,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )



import warnings
from typing import List, Union, Dict, Any, Tuple
from itertools import chain
import numpy as np

from .base import BoolElement, ElementBase, FloatElement, Pose, StringElement, vector3
from .sensor import Sensor
from .frame import Frame
from .origin import Origin
from .... import transform as tf


class Joint(ElementBase):
    """A constraint between two rigid bodys.

    A joint connects two links with kinematic and dynamic properties. This
    defines a constraint between the two bodies, and constructing a
    :mod``transform graph <skbot.transform>`` using ``to_dynamic_graph`` will
    build this constraint into the underlying graph structure.

    Parameters
    ----------
    name : str
        The name of the joint.
    type : str
        The type of joint, which must be one of the following:
            continuous
                A hinge joint that rotates around :class:``Joint.Axis``with a
                continuous range of motion.  This means that the upper
                and lower axis limits are ignored.
            revolute
                a hinge joint that rotates around :class:``Joint.Axis`` with
                a fixed range of motion.
            gearbox
                A geared revolute joints.
            revolute2
                Two revolute joints connected in series. The first one rotates
                around :class:``Joint.Axis``, and the second one rotates around
                :class:``Joint.Axis``.
            prismatic
                A sliding joint that slides along :class:``Joint.Axis`` with a
                limited range of motion.
            ball
                A ball and socket joint.
            screw
                A 1DoF joint with coupled sliding and rotational motion.
            universal
                Like a joint with type ``ball`` but constrains one degree of
                freedom.
            fixed
                A joint with zero DoF that rigidly connects two links.
    parent : Union[str, Joint.Parent]
        The first :class:`Link` that this joint constraints. It may be set to
        "world", in which case the movement of the rigid body in ``Joint.child``
        is constrained relative to the inertial world frame.

        .. versionchanged:: SDFormat v1.7
            Parent may now be set to "world" and, if so, will connect
            to the inertial world frame.
        .. versionchanged:: SDFormat v1.2
            Parent is now a string instead of :class:`Joint.Parent`.
    child : Union[str, Joint.Child]
        The second :class:`Link` that this joint constraints.

        .. versionchanged:: SDFormat v1.2
            Child is now a string instead of :class:`Joint.Child`.
    pose : Pose
        The links's initial position (x,y,z) and orientation (roll, pitch, yaw).

        .. versionadded:: SDFormat 1.2
    gearbox_ratio : float
        Parameter for gearbox joints. Given theta_1 and theta_2 defined in
        description for gearbox_reference_body, ``theta_2 = -gearbox_ratio *
        theta_1``. Default: 1.0

        .. versionadded:: SDFormat v1.4
    gearbox_reference_body : str
        Parameter for gearbox joints. Gearbox ratio is enforced over two joint
        angles. First joint angle (theta_1) is the angle from the
        gearbox_reference_body to the parent link defined by :class:`Joint.Axis`
        and the second joint angle (theta_2) is the angle from the
        gearbox_reference_body to the child link defined by
        :class:`Joint.Axis`.

        .. versionadded:: SDFormat v1.4
    thread_pitch : float
        Parameter for screw joints. The amount of linear displacement for each
        full rotation of the joint.
    axis : Joint.Axis
        Configuration parameters for joints that rotate around or translate
        along at least one axis.
    axis2 : Joint.Axis
        Configuration parameters for joints that rotate around or translate
        along at least two axis.
    physics : Joint.Physics
        Configuration parameters for various physics engines. (Currently: ODE
        and Simbody.)
    frames : List[Frame]
        A list of frames of reference in which poses may be expressed.

        .. deprecated:: SDFormat v1.7
            Use :attr:`Model.frame` instead.
        .. versionadded:: SDFormat v1.5
    sensors : List[Sensor]
        A list of sensors attached to this joint.

        .. versionadded:: SDFormat v1.4
    sdf_version : str
        The SDFormat version to use when constructing this element.
    origin : Origin
        The joint's origin.

        .. deprecated:: SDFormat v1.2
            Use ``Joint.pose`` instead.

    Attributes
    ----------
    name : str
        See ``Parameters`` section.
    type : str
        See ``Parameters`` section.
    parent : str
        See ``Parameters`` section.
    child : str
        See ``Parameters`` section.
    pose : Pose
        See ``Parameters`` section.
    gearbox_ratio : float
        See ``Parameters`` section.
    gearbox_reference_body : str
        See ``Parameters`` section.
    thread_pitch : float
        See ``Parameters`` section.
    axis : Joint.Axis
        See ``Parameters`` section.
    axis2 : Joint.Axis
        See ``Parameters`` section.
    physics : Joint.Physics
        See ``Parameters`` section.
    frames : List[Frame]
        See ``Parameters`` section.
    sensors : List[Sensor]
        See ``Parameters`` section.
    sdf_version : str
        See ``Parameters`` section.
    origin : Origin
        See ``Parameters`` section.

    Notes
    -----
    A joint defines an implicit frame of reference to which other
    :class:`Frame`s may be ``attached_to``. As such its name must be unique
    among all frames and frame-bearing elements.

    """

    def __init__(
        self,
        *,
        name: str,
        type: str,
        parent: Union[str, "Joint.Parent"],
        child: Union[str, "Joint.Child"],
        origin: Origin = None,
        pose: Pose = None,
        gearbox_ratio: float = 1.0,
        gearbox_reference_body: str = None,
        thread_pitch: float = 1.0,
        axis: "Joint.Axis" = None,
        axis2: "Joint.Axis" = None,
        physics: "Joint.Physics" = None,
        frames: List[Frame] = None,
        sensors: List[Sensor] = None,
        sdf_version: str,
    ) -> None:
        super().__init__(sdf_version=sdf_version)
        self.name = name
        self.type = type
        if origin is None:
            self._origin = Origin(sdf_version=sdf_version)
        elif sdf_version == "1.0":
            self._origin = origin
        else:
            warnings.warn("`origin` is deprecated. Use `pose` instead.")
            self._origin = origin
        if sdf_version == "1.0":
            self.pose = self._origin.pose
        elif pose is None:
            self.pose = Pose(sdf_version=sdf_version)
        else:
            self.pose = pose
        self.gearbox_ratio = gearbox_ratio
        self.gearbox_reference_body = gearbox_reference_body
        self.thread_pitch = thread_pitch
        self.axis = Joint.Axis(sdf_version=sdf_version) if axis is None else axis
        self.axis2 = Joint.Axis(sdf_version=sdf_version) if axis2 is None else axis2
        self.physics = (
            Joint.Physics(sdf_version=sdf_version) if physics is None else physics
        )
        self._frames = [] if frames is None else frames
        self.sensors = [] if sensors is None else sensors

        self._origin.pose = self.pose

        if isinstance(parent, Joint.Parent):
            self.parent = parent.link
        else:
            self.parent = parent

        if isinstance(child, Joint.Child):
            self.child = child.link
        else:
            self.child = child

        if self.axis._use_parent_model_frame:
            self.axis.xyz.expressed_in = self.parent
        elif self.axis.xyz.expressed_in is None:
            self.axis.xyz.expressed_in = self.child

        if self.axis2._use_parent_model_frame:
            self.axis2.xyz.expressed_in = self.parent
        elif self.axis2.xyz.expressed_in is None:
            self.axis2.xyz.expressed_in = self.child

        for frame in self._frames:
            if frame.attached_to is None:
                frame.attached_to = self.name
            if frame.pose.relative_to is None:
                frame.pose.relative_to = self.name

        for sensor in sensors:
            if sensor.pose.relative_to is None:
                sensor.pose.relative_to = self.name

            if sensor.camera.pose.relative_to is None:
                sensor.camera.pose.relative_to = f"{self.name}::{sensor.name}"

            for frame in sensor.camera._frames:
                if frame.pose.relative_to is None:
                    frame.pose.relative_to = (
                        f"{self.name}::{sensor.name}::{sensor.camera.name}"
                    )
                if frame.attached_to is None:
                    frame.attached_to = f"{self.name}::{sensor.name}"

            for frame in sensor._frames:
                if frame.pose.relative_to is None:
                    frame.pose.relative_to = f"{self.name}::{sensor.name}"

                if frame.attached_to is None:
                    frame.attached_to = f"{self.name}::{sensor.name}"

    @property
    def origin(self):
        warnings.warn(
            "`Joint.origin` is deprecated since SDFormat v1.2."
            " Use `Joint.pose` instead.",
            DeprecationWarning,
        )
        return self._origin

    @property
    def frames(self):
        warnings.warn(
            "`Link.frames` is deprecated since SDF v1.7."
            " Use `Model.frames` instead and set `Frame.attached_to`"
            " to the name of this joint.",
            DeprecationWarning,
        )
        return self._frames

    @classmethod
    def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
        joint_args = {
            "name": specific.name,
            "type": specific.type,
            "parent": specific.parent,
            "child": specific.child,
        }
        args_with_default = {
            "origin": Origin,
            "pose": Pose,
            "gearbox_ratio": FloatElement,
            "gearbox_reference_body": StringElement,
            "thread_pitch": FloatElement,
            "axis": Joint.Axis,
            "axis2": Joint.Axis,
            "physics": Joint.Physics,
        }
        list_args = {"frame": ("frames", Frame), "sensor": ("sensors", Sensor)}
        standard_args = cls._prepare_standard_args(
            specific, args_with_default, list_args, version=version
        )
        joint_args.update(standard_args)

        return Joint(**joint_args, sdf_version=version)

    def declared_frames(self) -> Dict[str, tf.Frame]:
        joint_frame = tf.Frame(3, name=self.name)
        declared_frames = {
            self.name: joint_frame,
            self.name + "_child": joint_frame,
            self.name + "_parent": tf.Frame(3, name=self.name + "_parent"),
        }

        for el in chain(self._frames):
            declared_frames.update(el.declared_frames())

        for sensor in self.sensors:
            for name, frame in sensor.declared_frames().items():
                declared_frames[f"{self.name}::{name}"] = frame

        return declared_frames

    def to_static_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        *,
        seed: int = None,
        shape: Tuple,
        axis: int = -1,
    ) -> tf.Frame:
        parent_name = self.pose.relative_to
        joint_child = self.name + "_child"
        joint_parent = self.name + "_parent"

        parent = declared_frames[parent_name]
        joint_child = declared_frames[joint_child]
        joint_parent = declared_frames[joint_parent]

        link = self.pose.to_tf_link()
        link(joint_child, parent)
        link(joint_parent, parent)

        for frame in self._frames:
            frame.to_static_graph(declared_frames, seed=seed, shape=shape, axis=axis)

        for sensor in self.sensors:
            sensor.to_static_graph(
                declared_frames,
                f"{self.name}::{sensor.name}",
                seed=seed,
                shape=shape,
                axis=axis,
            )

        return declared_frames[self.name]

    def to_dynamic_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        *,
        seed: int = None,
        shape: Tuple,
        axis: int = -1,
        apply_state: bool = True,
        _scaffolding: Dict[str, tf.Frame],
    ) -> tf.Frame:
        parent_name = self.parent
        child_name = self.child
        joint_name = self.name
        joint_name_child = self.name + "_child"
        joint_name_parent = self.name + "_parent"

        parent = declared_frames[parent_name]
        child = declared_frames[child_name]
        joint_parent = declared_frames[joint_name_parent]
        joint_child = declared_frames[joint_name_child]

        parent_static = _scaffolding[parent_name]
        joint_static = _scaffolding[joint_name]
        child_static = _scaffolding[child_name]

        link = tf.CompundLink(parent_static.links_between(joint_static))
        link(parent, joint_parent)

        joint_link: tf.Link = self.to_tf_link(_scaffolding, shape=shape, axis=axis)
        joint_link(joint_child, joint_parent)

        link = tf.CompundLink(joint_static.links_between(child_static))
        link(joint_child, child)

        for el in chain(self._frames):
            el.to_dynamic_graph(
                declared_frames,
                seed=seed,
                shape=shape,
                axis=axis,
                apply_state=apply_state,
                _scaffolding=_scaffolding,
            )

        for sensor in self.sensors:
            parent_name = self.name
            child_name = f"{self.name}::{sensor.name}"

            parent = declared_frames[parent_name]
            child = declared_frames[child_name]

            parent_static = _scaffolding[parent_name]
            child_static = _scaffolding[child_name]

            link = tf.CompundLink(parent_static.links_between(child_static))
            link(parent, child)

            sensor.to_dynamic_graph(
                declared_frames,
                f"{self.name}::{sensor.name}",
                seed=seed,
                shape=shape,
                axis=axis,
                apply_state=apply_state,
                _scaffolding=_scaffolding,
            )

        return declared_frames[self.name]

    def to_tf_link(
        self, scaffolding: Dict[str, tf.Frame], *, shape: Tuple = (3,), axis=-1
    ) -> tf.Link:
        """tf.Link from **child** to **parent** frame

        Parameters
        ----------
        scaffolding : Dict[str, tf.Frame]
            A static graph that allows the resolution of ``Joint.Axis.Xyz.expressed_in``
            (``Joint.Axis.Xyz.Value`` needs to be converted to the the implicit frame of
            ``Joint.child`` as per the spec.)
        shape : tuple
            A tuple describing the shape of elements that the resulting graph should
            transform. This can be used to add batch dimensions to the graph, for
            example to perform vectorized computation on multiple instances of the
            same world in different states, or for batched coordinate
            transformation. Defaults to (3,), which is the shape of a single 3D
            vector in euclidian space.
        axis : int
            The axis along which elements are stored. The axis must have length 3
            (since SDFormat describes 3 dimensional worlds), and all other axis are
            considered batch dimensions. Defaults to -1.

        Returns
        -------
        joint_link : tf.Link
            A parameterized link that represents transformation from the joint's
            child into the joint's parent using the constraint defined by this
            joint.

        """
        expressed_frame = scaffolding[self.axis.xyz.expressed_in]
        child_frame = scaffolding[self.child]
        joint_axis1 = expressed_frame.transform(self.axis.xyz.value, child_frame)

        if self.type == "continuous":
            warnings.warn("Hinge joints have not been added yet.")
            link = tf.Translation((0, 0, 0))
        elif self.type == "revolute":
            angle = np.clip(0, self.axis.limit.lower, self.axis.limit.upper)
            link = tf.RotationalJoint(
                joint_axis1,
                angle=angle,
                upper_limit=self.axis.limit.upper,
                lower_limit=self.axis.limit.lower,
                axis=axis,
            )
        elif self.type == "hinge":
            warnings.warn("Hinge joints have not been added yet.")
            link = tf.Translation((0, 0, 0))
        elif self.type == "gearbox":
            warnings.warn("Gearbox joints have not been added yet.")
            link = tf.Translation((0, 0, 0))
        elif self.type == "revolute2":
            warnings.warn("Revolute2 type joint is not added yet.")
            link = tf.Translation((0, 0, 0))
        elif self.type == "prismatic":
            amount = np.clip(1, self.axis.limit.lower, self.axis.limit.upper)
            link = tf.PrismaticJoint(
                joint_axis1,
                amount=amount,
                upper_limit=self.axis.limit.upper,
                lower_limit=self.axis.limit.lower,
                axis=axis,
            )
        elif self.type == "ball":
            warnings.warn("Ball joints have not been added yet.")
            link = tf.Translation((0, 0, 0))
        elif self.type == "screw":
            warnings.warn("Screw joints have not been added yet.")
            link = tf.Translation((0, 0, 0))
        elif self.type == "universal":
            warnings.warn("Universal joints have not been added yet.")
            link = tf.Translation((0, 0, 0))
        elif self.type == "fixed":
            link = tf.Translation((0, 0, 0))
        else:
            raise ValueError(f"Unknown Joint type: {self.type}")

        return link

    class Parent(ElementBase):
        def __init__(self, *, link: str, sdf_version: str) -> None:
            super().__init__(sdf_version=sdf_version)
            self.link = link

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
            return Joint.Parent(link=specific.link, version=version)

    class Child(ElementBase):
        def __init__(self, *, link: str, sdf_version: str) -> None:
            super().__init__(sdf_version=sdf_version)
            self.link = link

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
            return Joint.Parent(link=specific.link, version=version)

    class Axis(ElementBase):
        """Parameters for rotating and/or translating joints.

        Parameters
        ----------
        initial_position : float
            Default joint position for this joint axis.

            .. deprecated:: SDFormat v1.8
                Use `World.State` instead to set the initial position.
            .. versionadded:: SDFormat v1.6
        xyz: Union[str, Joint.Axis.Xyz]
            The direction of the Axis.

            .. versionchanged:: SDFormat v1.7
                ``xyz`` is now a class. The direction vector is stored
                in ``xyz.value``.
        use_parent_model_frame : bool
            Flag to interpret the axis xyz element in the parent model frame
            instead of joint frame. Provided for Gazebo compatibility (see
            https://github.com/osrf/gazebo/issue/494 ). Default is: ``False``.

            .. deprecated:: SDFormat v1.7
                Use :attr:`Joint.Axis.Xyz.expressed_in` instead.
            .. versionadded:: SDFormat v1.5
        dynamics: Joint.Axis.Dynamics
            Dynamic Parameters related to the Axis.
        limit : Joint.Axis.Limit
            Constraints applied to parameters of this joint, e.g., upper/lower
            limits.

            .. versionchanged:: SDFormat v1.6
                The limit element is now optional.
        sdf_version : str
            The SDFormat version to use when constructing this element.

        Attributes
        ----------
        initial_position : float
            See ``Parameters`` section.
        xyz: Joint.Axis.Xyz
            See ``Parameters`` section.
        dynamics: Joint.Axis.Dynamics
            See ``Parameters`` section.
        limit : Joint.Axis.Limit
            See ``Parameters`` section.

        """

        def __init__(
            self,
            *,
            initial_position: float = 0,
            xyz: Union[str, "Joint.Axis.Xyz"] = None,
            use_parent_model_frame: bool = False,
            dynamics: "Joint.Axis.Dynamics" = None,
            limit: "Joint.Axis.Limit" = None,
            sdf_version: str,
        ) -> None:
            super().__init__(sdf_version=sdf_version)
            self.initial_position = initial_position
            if sdf_version in ["1.0", "1.2", "1.3", "1.4", "1.5", "1.6"]:
                if xyz is None:
                    # old default
                    xyz = "0 0 1"
                self.xyz = Joint.Axis.Xyz(
                    value=xyz, expressed_in=None, sdf_version=sdf_version
                )
            elif xyz is None:
                self.xyz = Joint.Axis.Xyz(sdf_version=sdf_version)
            else:
                self.xyz = xyz
            self._use_parent_model_frame = use_parent_model_frame
            self.dynamics = (
                Joint.Axis.Dynamics(sdf_version=sdf_version)
                if dynamics is None
                else dynamics
            )
            self.limit = (
                Joint.Axis.Limit(sdf_version=sdf_version) if limit is None else limit
            )

        @property
        def use_parent_model_frame(self):
            warnings.warn(
                "`Joint.Axis.use_parent_model_frame` is deprecated."
                " Use `Joint.Axis.Xyz.expressed_in` instead."
            )
            return self._use_parent_model_frame

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
            args_with_default = {
                "initial_position": FloatElement,
                "xyz": Joint.Axis.Xyz,
                "use_parent_model_Frame": BoolElement,
                "dynamics": Joint.Axis.Dynamics,
                "limit": Joint.Axis.Limit,
            }

            if version in ["1.0", "1.2", "1.3", "1.4", "1.5", "1.6"]:
                args_with_default["xyz"] = StringElement

            standard_args = cls._prepare_standard_args(
                specific, args_with_default, version=version
            )

            return Joint.Axis(**standard_args, sdf_version=version)

        class Xyz(ElementBase):
            """The direction of an axis.

            Represents the x,y,z components of the axis unit vector. The
            vector should be normalized.

            .. versionadded:: SDFormat v1.7

            Parameters
            ----------
            value : str
                The numerical value of the direction. Default: "0 0 1".
            expressed_in : str
                The frame of reference in which the direction is expressed.
            sdf_version : str
                The SDFormat version to use when constructing this element.

            Attributes
            ----------
            value : np.ndarray
                See ``Parameters`` section.
            expressed_in : str
                See ``Parameters`` section.

            """

            def __init__(
                self,
                *,
                value: str = "0 0 1",
                expressed_in: str = None,
                sdf_version: str,
            ) -> None:
                super().__init__(sdf_version=sdf_version)
                self.value = vector3(value)

                self.expressed_in = expressed_in
                if self.expressed_in == "":
                    self.expressed_in = None

            @classmethod
            def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
                return Joint.Axis.Xyz(
                    value=specific.value,
                    expressed_in=specific.expressed_in,
                    sdf_version=version,
                )

        class Dynamics(ElementBase):
            """Dynamic Parameters along an axis.

            An element specifying physical properties of the joint related to
            the axis. These values are used to specify modeling properties of
            the joint, particularly useful for simulation.

            Parameters
            ----------
            damping : float
                Viscous damping coefficient along this axis. The default is
                ``0``.
            friction : float
                Static friction along this axis. The default is ``0``.
            spring_reference : float
                The neutral position of the spring along this axis. The default
                is ``0``.

                .. versionadded:: SDFormat v1.5
            spring_stiffness : float
                The stiffness of the spring along this axis. The default is
                ``0``.

                .. versionadded:: SDFormat v1.5
            sdf_version : str
                The SDFormat version to use when constructing this element. The
                default is ``0``.

            Attributes
            ----------
            damping : float
                See ``Parameters`` section.
            friction : float
                See ``Parameters`` section.
            spring_reference : float
                See ``Parameters`` section.
            spring_stiffness : float
                See ``Parameters`` section.
            """

            def __init__(
                self,
                *,
                damping: float = 0,
                friction: float = 0,
                spring_reference: float = 0,
                spring_stiffness: float = 0,
                sdf_version: str,
            ) -> None:
                super().__init__(sdf_version=sdf_version)
                self.damping = damping
                self.friction = friction
                self.spring_reference = spring_reference
                self.spring_stiffness = spring_stiffness

            @classmethod
            def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
                args_with_default = {
                    "damping": FloatElement,
                    "friction": FloatElement,
                    "spring_reference": FloatElement,
                    "spring_stiffness": FloatElement,
                }
                standard_args = cls._prepare_standard_args(
                    specific, args_with_default, version=version
                )
                return Joint.Axis.Dynamics(**standard_args, sdf_version=version)

        class Limit(ElementBase):
            """Joint limits along an axis.

            This class specifies limites/constraints for joint parameters along
            the containing axis.

            Parameters
            ----------
            lower : float
                A lower limit for any joint parameter along this axis. In
                radians for revolute joints and in meters for prismatic joints.
                Default: ``-1e16``.
            upper : float
                A upper limit for any joint parameter along this axis. In
                radians for revolute joints and in meters for prismatic joints.
                Default: ``1e16``.
            effort : float
                The maximum effort that may be applied to/by a joint along this
                axis. This limit is not enforced if its value is negative.
                Default: ``-1``.

                .. versionchanged:: SDFormat v1.3
                    The default changed from 0 to -1.
            velocity : float
                The maximum velocity (angular or directional respectively) that
                may be applied to/by this joint. This limit is not enforced if
                its value is negative. Default: ``-1``.

                .. versionchanged:: SDFormat v1.3
                    The default changed from 0 to -1.
            stiffness : float
                Joint stop stiffness. Default: ```1e8``.

                .. versionadded:: SDFormat v1.4
            dissipation : float
                Joint stop dissipation. Default: ```1.0``.

                .. versionadded:: SDFormat v1.4
            sdf_version : str
                The SDFormat version to use when constructing this element.

            Attributes
            ----------
            lower : float
                See ``Parameters`` section.
            upper : float
                See ``Parameters`` section.
            effort : float
                See ``Parameters`` section.
            velocity : float
                See ``Parameters`` section.
            stiffness : float
                See ``Parameters`` section.
            dissipation : float
                See ``Parameters`` section.

            """

            def __init__(
                self,
                *,
                lower: float = -1e16,
                upper: float = 1e16,
                effort: float = -1,
                velocity: float = -1,
                stiffness: float = 1e8,
                dissipation: float = 1.0,
                sdf_version: str,
            ) -> None:
                super().__init__(sdf_version=sdf_version)
                self.lower = lower
                self.upper = upper
                if effort < 0:
                    self.effort = float("inf")
                else:
                    self.effort = effort
                if velocity < 0:
                    self.velocity = float("inf")
                else:
                    self.velocity = velocity
                self.stiffness = stiffness
                self.dissipation = dissipation

            @classmethod
            def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
                args_with_default = {
                    "lower": FloatElement,
                    "upper": FloatElement,
                    "effort": FloatElement,
                    "velocity": FloatElement,
                    "stiffness": FloatElement,
                    "dissipation": FloatElement,
                }
                standard_args = cls._prepare_standard_args(
                    specific, args_with_default, version=version
                )
                return Joint.Axis.Limit(**standard_args, sdf_version=version)

    class Physics(ElementBase):
        """
          <element name="physics" required="0">
          <description>Parameters that are specific to a certain physics engine.</description>
          <element name="simbody" required="0">
            <description>Simbody specific parameters</description>
            <element name="must_be_loop_joint" type="bool" default="false" required="0">
              <description>Force cut in the multibody graph at this joint.</description>
            </element>
          </element>
          <element name="ode" required="0">
            <description>ODE specific parameters</description>
            <element name="cfm_damping" type="bool" default="false" required="0">
              <description>If cfm damping is set to true, ODE will use CFM to simulate damping, allows for infinite damping, and one additional constraint row (previously used for joint limit) is always active.</description>
            </element>

            <element name="implicit_spring_damper" type="bool" default="false" required="0">
              <description>If implicit_spring_damper is set to true, ODE will use CFM, ERP to simulate stiffness and damping, allows for infinite damping, and one additional constraint row (previously used for joint limit) is always active.  This replaces cfm_damping parameter in SDFormat 1.4.</description>
            </element>

            <element name="fudge_factor" type="double" default="0" required="0">
              <description>Scale the excess for in a joint motor at joint limits. Should be between zero and one.</description>
            </element>
            <element name="cfm" type="double" default="0" required="0">
              <description>Constraint force mixing for constrained directions</description>
            </element>
            <element name="erp" type="double" default="0.2" required="0">
              <description>Error reduction parameter for constrained directions</description>
            </element>
            <element name="bounce" type="double" default="0" required="0">
              <description>Bounciness of the limits</description>
            </element>
            <element name="max_force" type="double" default="0" required="0">
              <description>Maximum force or torque used to reach the desired velocity.</description>
            </element>
            <element name="velocity" type="double" default="0" required="0">
              <description>The desired velocity of the joint. Should only be set if you want the joint to move on load.</description>
            </element>

            <element name="limit" required="0">
              <description></description>
              <element name="cfm" type="double" default="0.0" required="1">
                <description>Constraint force mixing parameter used by the joint stop</description>
              </element>
              <element name="erp" type="double" default="0.2" required="1">
                <description>Error reduction parameter used by the joint stop</description>
              </element>
            </element>

            <element name="suspension" required="0">
              <description></description>
              <element name="cfm" type="double" default="0.0" required="1">
                <description>Suspension constraint force mixing parameter</description>
              </element>
              <element name="erp" type="double" default="0.2" required="1">
                <description>Suspension error reduction parameter</description>
              </element>
            </element>
          </element>

          <element name="provide_feedback" type="bool" default="false" required="0">
            <description>If provide feedback is set to true, physics engine will compute the constraint forces at this joint.</description>
          </element>
        </element> <!-- End Physics -->
        """

        def __init__(
            self,
            *,
            simbody: "Joint.Physics.Simbody" = None,
            ode: "Joint.Physics.Ode" = None,
            provide_feedback: bool = False,
            sdf_version: str,
        ) -> None:
            warnings.warn("`Joint.Physics` is not implemented yet.")
            super().__init__(sdf_version=sdf_version)

        class Simbody(ElementBase):
            def __init__(
                self,
                *,
                must_be_loop_joint: bool = False,
                sdf_version: str,
            ) -> None:
                warnings.warn("`Joint.Physics.Simbody` is not implemented yet.")
                super().__init__(sdf_version=sdf_version)

        class Ode(ElementBase):
            def __init__(
                self,
                *,
                provide_feedback: bool = False,
                cfm_damping: bool = False,
                implicit_spring_damper: bool = False,
                fudge_factor: float = 0,
                cfm: float = 0,
                erp: float = 0.2,
                bounce: float = 0,
                max_force: float = 0,
                velocity: float = 0,
                limit: "Joint.Physics.Ode.Limit" = None,
                suspension: "Joint.Physics.Ode.Suspension" = None,
                sdf_version: str,
            ) -> None:
                warnings.warn("`Joint.Physics.Ode` is not implemented yet.")
                super().__init__(sdf_version=sdf_version)

            class Limit(ElementBase):
                def __init__(
                    self,
                    *,
                    cfm: float = 0,
                    erp: float = 0.2,
                    sdf_version: str,
                ) -> None:
                    warnings.warn("`Joint.Physics.Ode.Limit` is not implemented yet.")
                    super().__init__(sdf_version=sdf_version)

            class Suspension(ElementBase):
                def __init__(
                    self,
                    *,
                    cfm: float = 0,
                    erp: float = 0.2,
                    sdf_version: str,
                ) -> None:
                    warnings.warn(
                        "`Joint.Physics.Ode.Suspension` is not implemented yet."
                    )
                    super().__init__(sdf_version=sdf_version)



import warnings
from typing import DefaultDict, List, Dict, Any, Tuple, Union

from .base import ElementBase, FloatElement, IntegerElement, Pose, StringElement
from .frame import Frame
from .... import transform as tf
from ...transformations import FrustumProjection


class Camera(ElementBase):
    """A Camera Sensor.

    This element describes the parameters of a perspective camera system with
    optional lense, noise, and distortion effects. Contrary to common
    formulations the camera's direction of view is along the x-axis and the
    image plane runs parallel to the y-z-plane. If u and v describe the two axis
    of the of the image plane, then the u-axis runs along the negative y axis,
    and the v-axis runs along the negative z-axis. The origin of the image plane
    is along the top left corner of the viewing frustum.

    Parameters
    ----------
    name : str
        The name of the camera. Default: ``camera``

        .. versionadded:: SDFormat 1.3
    pose : Pose
        The camera's initial position (x,y,z) and orientation (roll, pitch, yaw).

        .. versionadded:: SDFormat 1.3
    horizontal_fov : Union[float, Camera.HorizontalFov]
        The viewing frustum's horizontal angle of view along the image's u-axis or world's y-axis.
        Default: ``1.047``.

        .. versionchanged:: SDFormat 1.2
            horizontal_fov is now a float instead of a class.
    image : Camera.Image
        Layout of the captured images (shape and color format).
    clip : Camera.Clip
        Near and Far clip plane of the camera.
    save : Camera.Save
        Configuration parameters related to saving rendered images.
    depth_camera : Camera.DepthCamera
        Configuration parameters related to the depth camera.
    noise : Camera.Noise
        Parameters for the noise model.

        .. versionadded:: SDFormat 1.4
    distortion : Camera.Distortion
        Parameters for the distortion model.

        .. versionadded:: SDFormat 1.5
    lense : Camera.Lense
        Parameters for the lense model.

        .. versionadded:: SDFormat 1.5
    visibility_mask: int
        Visibility mask of a camera. When (camera's visibility_mask & visual's
        visibility_flags) evaluates to non-zero, the visual will be visible to
        the camera.

        .. versionadded:: SDFormat 1.7
    sdf_version : str
        The SDFormat version to use when constructing this element.
    frames : List[Frame]
        A list of frames of reference in which poses may be expressed.

        .. deprecated:: SDFormat v1.7
            Use :attr:`Model.frame` instead.
        .. versionadded:: SDFormat v1.5

    Attributes
    ----------
    name : str
        See ``Parameters`` section.
    pose : Pose
        See ``Parameters`` section.
    horizontal_fov : Union[float, Camera.HorizontalFov]
        See ``Parameters`` section.
    image : Camera.Image
        See ``Parameters`` section.
    clip : Camera.Clip
        See ``Parameters`` section.
    save : Camera.Save
        See ``Parameters`` section.
    depth_camera : Camera.DepthCamera
        See ``Parameters`` section.
    noise : Camera.Noise
        See ``Parameters`` section.
    distortion : Camera.Distortion
        See ``Parameters`` section.
    lense : Camera.Lense
        See ``Parameters`` section.
    visibility_mask: int
        See ``Parameters`` section.


    Notes
    -----
    The frustum's vertical angle of view (along the v-axis / z-axis) chosen
    using ``horizontal_fov`` and the pixel dimension, i.e., the aspect ratio, of
    the image.

    """

    def __init__(
        self,
        *,
        name: str = "camera",
        pose: Pose = None,
        horizontal_fov: Union[float, "Camera.HorizontalFov"] = 1.047,
        image: "Camera.Image" = None,
        clip: "Camera.Clip" = None,
        save: "Camera.Save" = None,
        depth_camera: "Camera.DepthCamera" = None,
        noise: "Camera.Noise" = None,
        distortion: "Camera.Distortion" = None,
        lense: "Camera.Lense" = None,
        frames: List[Frame] = None,
        visibility_mask: int = 4294967295,
        sdf_version: str,
    ) -> None:
        warnings.warn("`Camera` has not been implemented yet.")
        super().__init__(sdf_version=sdf_version)

        self.name = name
        self.pose = Pose(sdf_version=sdf_version) if pose is None else pose
        if sdf_version == "1.0":
            self.horizontal_fov = (
                Camera.HorizontalFov(sdf_version=sdf_version)
                if horizontal_fov is None
                else horizontal_fov
            )
        else:
            self.horizontal_fov = horizontal_fov
        self.image = Camera.Image(sdf_version=sdf_version) if image is None else image
        self.clip = Camera.Clip(sdf_version=sdf_version) if clip is None else clip
        self.save = Camera.Save(sdf_version=sdf_version) if save is None else save
        self.depth_camera = (
            Camera.DepthCamera(sdf_version=sdf_version)
            if depth_camera is None
            else depth_camera
        )
        self.noise = Camera.Noise(sdf_version=sdf_version) if noise is None else noise
        self.distortion = (
            Camera.Distortion(sdf_version=sdf_version)
            if distortion is None
            else distortion
        )
        self.lense = Camera.Lense(sdf_version=sdf_version) if lense is None else lense
        self._frames = [] if frames is None else frames
        self.visibility_mask = visibility_mask

    @property
    def frames(self):
        warnings.warn(
            "`Sensor.frames` is deprecated since SDF v1.7."
            " Use `Model.frames` instead and set `Frame.attached_to` to the name of this link.",
            DeprecationWarning,
        )
        return self._frames

    @classmethod
    def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
        # camera_args = {"visibility_mask": specific.visibility_mask}

        default_args = {
            "name": StringElement,
            "pose": Pose,
            "image": Camera.Image,
            "clip": Camera.Clip,
            "save": Camera.Save,
            "depth_camera": Camera.DepthCamera,
            "noise": Camera.Noise,
            "distortion": Camera.Distortion,
            "lense": Camera.Lense,
            "visbility_mask": IntegerElement,
        }
        if version == "1.0":
            default_args["horizontal_fov"] = Camera.HorizontalFov
        else:
            default_args["horizontal_fov"] = FloatElement
        list_args = {"frame": ("frames", Frame)}
        standard_args = cls._prepare_standard_args(
            specific, default_args, list_args, version=version
        )

        return Camera(**standard_args, sdf_version=version)

    def declared_frames(self) -> Dict[str, tf.Frame]:
        declared_frames = {
            self.name: tf.Frame(3, name=self.name),
            "pixel_space": tf.Frame(2, name="pixel_space"),
        }

        for frame in self._frames:
            declared_frames.update(frame.declared_frames())

        return declared_frames

    def to_static_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        sensor_frame: str,
        *,
        seed: int = None,
        shape: Tuple,
        axis: int = -1,
    ) -> tf.Frame:
        self.pose.to_static_graph(
            declared_frames, sensor_frame + f"::{self.name}", shape=shape, axis=axis
        )

        if self.sdf_version == "1.0":
            hfov = self.horizontal_fov.angle
        else:
            hfov = self.horizontal_fov

        parent = declared_frames[sensor_frame + f"::{self.name}"]
        child = declared_frames[sensor_frame + "::pixel_space"]
        projection = FrustumProjection(hfov, (self.image.height, self.image.width))
        projection(parent, child)

        for frame in self._frames:
            frame.pose.to_static_graph(
                declared_frames,
                sensor_frame + f"::{frame.name}",
                shape=shape,
                axis=axis,
            )

        return declared_frames[sensor_frame + f"::{self.name}"]

    def to_dynamic_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        sensor_frame: str,
        *,
        seed: int = None,
        shape: Tuple,
        axis: int = -1,
        apply_state: bool = True,
        _scaffolding: Dict[str, tf.Frame],
    ) -> tf.Frame:
        parent_name = sensor_frame
        child_name = sensor_frame + f"::{self.name}"

        parent = declared_frames[parent_name]
        child = declared_frames[child_name]

        parent_static = _scaffolding[parent_name]
        child_static = _scaffolding[child_name]

        link = tf.CompundLink(parent_static.links_between(child_static))
        link(parent, child)

        if self.sdf_version == "1.0":
            hfov = self.horizontal_fov.angle
        else:
            hfov = self.horizontal_fov

        parent = declared_frames[sensor_frame + f"::{self.name}"]
        child = declared_frames[sensor_frame + "::pixel_space"]
        projection = FrustumProjection(hfov, (self.image.height, self.image.width))
        projection(parent, child)

        return declared_frames[sensor_frame + f"::{self.name}"]

    class HorizontalFov(ElementBase):
        def __init__(self, *, angle: float = 1.047, sdf_version: str) -> None:
            super().__init__(sdf_version=sdf_version)
            self.angle = angle

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
            return Camera.HorizontalFov(angle=specific.angle, sdf_version=version)

    class Image(ElementBase):
        """Image shape and color format

        An image is a rectangular subset of a 2D plane measured in pixels.
        Coordinates on this plane are referred to using a two dimensional
        coordinate system with basis vectors u and v (the uv-plane).

        Parameters
        ----------
        width : int
            Number of pixels along the v-axis. Default: ``320``.
        height : int
            Number of pixels along the u-axis. Default: ``240``.
        format : str
            The data format of the color channel. This value determines the
            number of channels, their order, and their meaning. Possible values
            are: L8, L16, R_FLOAT16, R_FLOAT32, R8G8B8, B8G8R8, BAYER_RGGB8,
            BAYER_BGGR8, BAYER_GBRG8, BAYER_GRBG8. Default: ``R8G8B8``.

            .. versionadded:: SDFormat v1.7
                Formats: L16, R_FLOAT16, R_FLOAT32
        sdf_version : str
            The SDFormat version to use when constructing this element.

        Attributes
        ----------
        width : int
            See ``Parameters`` section.
        height : int
            See ``Parameters`` section.
        format : str
            See ``Parameters`` section.

        """

        def __init__(
            self,
            *,
            width: int = 320,
            height: int = 240,
            format: str = "R8G8B8",
            sdf_version: str,
        ) -> None:
            super().__init__(sdf_version=sdf_version)
            self.width = width
            self.height = height
            self.format = format

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
            return Camera.Image(
                width=specific.width,
                height=specific.height,
                format=specific.format,
                sdf_version=version,
            )

    class Clip(ElementBase):
        """Near and Far Clipping distance.

        Parameters
        ----------
        near : float
            Near clip distance. Default: ``0.1``.
        far : float
            Far clip distance. Default: ``100``.
        sdf_version : str
            The SDFormat version to use when constructing this element.

        Attributes
        ----------
        near : float
            See ``Parameters`` section.
        far : float
            See ``Parameters`` section.

        """

        def __init__(
            self, *, near: float = 0.1, far: float = 100, sdf_version: str
        ) -> None:
            super().__init__(sdf_version=sdf_version)
            self.near = near
            self.far = far

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
            return Camera.Clip(
                near=specific.near, far=specific.far, sdf_version=version
            )

    class Save(ElementBase):
        """Save/Export frames as images.

        Parameters
        ----------
        enabled : bool
            Enable export as images. Default: ``False``.
        path : str
            The path name which will hold the frame data. If path name is
            relative, then directory is relative to current working directory.
            Default: None
        sdf_version : str
            The SDFormat version to use when constructing this element.

        Attributes
        ----------
        enabled : bool
            See ``Parameters`` section.
        path : str
            See ``Parameters`` section.

        """

        def __init__(
            self, *, enabled: bool = False, path: str = None, sdf_version: str
        ) -> None:
            super().__init__(sdf_version=sdf_version)
            self.enabled = enabled
            self.path = path

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
            return Camera.Save(
                enabled=specific.enabled, path=specific.path, sdf_version=version
            )

    class DepthCamera(ElementBase):
        """Depth Camera configuration

        Parameters
        ----------
        output : str
            The type of data to output. Default: ``depths``.
        clip: Camera.Clip
            Depth camera clipping distance.
        sdf_version : str
            The SDFormat version to use when constructing this element.

        Attributes
        ----------
        output : str
            See ``Parameters`` section.
        clip: Camera.Clip
            See ``Parameters`` section.

        """

        def __init__(
            self,
            *,
            output: str = "depths",
            clip: "Camera.Clip" = None,
            sdf_version: str,
        ) -> None:
            super().__init__(sdf_version=sdf_version)
            self.output = output
            self.clip = clip

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
            args_with_default = {"clip": Camera.Clip}
            standard_args = cls._prepare_standard_args(
                specific, args_with_default, version=version
            )
            return Camera.DepthCamera(
                output=specific.output, **standard_args, sdf_version=version
            )

    class Noise(ElementBase):
        """The camera's noise model.

        Parameters
        ----------
        type : str
            The shape of the random noise distribution. Currently supported
            types are: "gaussian" (draw additive noise values independently for
            each pixel from a Gaussian distribution). Default: ``gaussian``.
        mean : float
            The distribution mean. Default: ``0``.
        stddev : float
            The distribution's standard deviation. Default: ``0``.
        sdf_version : str
            The SDFormat version to use when constructing this element.

        Attributes
        ----------
        type : str
            See ``Parameters`` section.
        mean : float
            See ``Parameters`` section.
        stddev : float
            See ``Parameters`` section.

        """

        def __init__(
            self,
            *,
            type: str = "gaussian",
            mean: float = 0,
            stddev: float = 0,
            sdf_version: str,
        ) -> None:
            super().__init__(sdf_version=sdf_version)
            self.type = type
            self.mean = (mean,)
            self.stddev = stddev

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
            return Camera.Noise(
                type=specific.type,
                mean=specific.mean,
                stddev=specific.stddev,
                sdf_version=version,
            )

    class Distortion(ElementBase):
        """The camera's distortion model

        Lens distortion to be applied to camera images. See
        http://en.wikipedia.org/wiki/Distortion_(optics)#Software_correction

        Parameters
        ----------
        k1 : float
            The first radial distortion coefficient. Default: ``0``.
        k2 : float
            The second radial distortion coefficient. Default: ``0``.
        k3 : float
            The third radial distortion coefficient. Default: ``0``.
        p1 : float
            The first tangential distortion coefficient. Default: ``0``
        p2 : float
            The second tangential distortion coefficient. Default: ``0``.
        center : str
            The distortion center or principal point. Default: ``0.5, 0.5``.
        sdf_version : str
            The SDFormat version to use when constructing this element.

        Attributes
        ----------
        k1 : float
            See ``Parameters`` section.
        k2 : float
            See ``Parameters`` section.
        k3 : float
            See ``Parameters`` section.
        p1 : float
            See ``Parameters`` section.
        p2 : float
            See ``Parameters`` section.
        center : str
            See ``Parameters`` section.

        """

        def __init__(
            self,
            *,
            k1: float = 0,
            k2: float = 0,
            k3: float = 0,
            p1: float = 0,
            p2: float = 0,
            center: str = "0.5, 0.5",
            sdf_version: str,
        ) -> None:
            super().__init__(sdf_version=sdf_version)
            self.k1 = k1
            self.k2 = k2
            self.k3 = k3
            self.p1 = p1
            self.p2 = p2
            self.center = center

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
            return Camera.Distortion(
                k1=specific.k1,
                k2=specific.k2,
                k3=specific.k3,
                p1=specific.p1,
                p2=specific.p2,
                center=specific.center,
                sdf_version=version,
            )

    class Lense(ElementBase):
        """A Camera's lense properties.

        Parameters
        ----------
        type : str
            Type of the lens mapping. Supported values are gnomonical,
            stereographic, equidistant, equisolid_angle, orthographic, custom.
            For gnomonical (perspective) projection, it is recommended to
            specify a horizontal_fov of less than or equal to 90°. Default:
            ``stereographic``.
        scale_to_hfov : bool
            If true the image will be scaled to fit horizontal FOV, otherwise it
            will be shown according to projection type parameters. Default:
            ``True``.
        custom_function : Camera.Lense.CustomFunction
            A custom lense map.
        cuttoff_angle : float
            Everything outside of the specified angle will be hidden. Default:
            ``.5707``.
        env_texture_size : int
            Resolution of the environment cube map used to draw the world.
            Default: ``56``.
        intrinsics : Camera.Lense.Intrinsics
            Parameters of a custom perspective matrix.
        sdf_version : str
            The SDFormat version to use when constructing this element.

        Attributes
        ----------
        type : str
            See ``Parameters`` section.
        scale_to_hfov : bool
            See ``Parameters`` section.
        custom_function : Camera.Lense.CustomFunction
            See ``Parameters`` section.
        cuttoff_angle : float
            See ``Parameters`` section.
        env_texture_size : int
            See ``Parameters`` section.
        intrinsics : Camera.Lense.Intrinsics
            See ``Parameters`` section.

        """

        def __init__(
            self,
            *,
            type: str = "stereographic",
            scale_to_hfov: bool = True,
            custom_function: "Camera.Lense.CustomFunction" = None,
            cuttoff_angle: float = 1.5707,
            env_texture_size: int = 256,
            intrinsics: "Camera.Lense.Intrinsics" = None,
            sdf_version: str,
        ) -> None:
            super().__init__(sdf_version=sdf_version)
            self.type = type
            self.scale_to_hfov = scale_to_hfov
            self.custom_function = (
                Camera.Lense.CustomFunction(sdf_version=sdf_version)
                if custom_function is None
                else custom_function
            )
            self.cuttoff_angle = cuttoff_angle
            self.env_texture_size = env_texture_size
            self.intrinsics = (
                Camera.Lense.Intrinsics(sdf_version=sdf_version)
                if intrinsics is None
                else intrinsics
            )

        @classmethod
        def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
            lense_args = {
                "type": specific.type,
                "scale_to_hfov": specific.scale_to_hfov,
                "custom_function": specific.custom_function,
                "cuttoff_angle": specific.cuttoff_angle,
                "env_texture_size": specific.env_texture_size,
            }
            args_with_default = {
                "custom_function": Camera.Lense.CustomFunction,
                "intrinsics": Camera.Lense.Intrinsics,
            }
            standard_args = cls._prepare_standard_args(
                specific, args_with_default, version=version
            )

            return Camera.Lense(**lense_args, **standard_args)

        class CustomFunction(ElementBase):
            """A custom lense map.

            Definition of custom mapping function in a form of
            r=c1*f*fun(theta/c2 + c3). See
            https://en.wikipedia.org/wiki/Fisheye_lens#Mapping_function

            Parameters
            ----------
            c1 : float
                Linear scaling constant. Default: ``1``.
            c2 : float
                Angle scaling constant. Default: ``1``.
            c3 : float
                Angle offset constant. Default: ``0``.
            f : float
                Focal length of the optical system. Note: It's not a focal
                length of the lens in a common sense! This value is ignored if
                'scale_to_fov' is set to true. Default: ``1``.
            fun : str
                The non-linear element to use. Possible values are 'sin', 'tan'
                and 'id'. Default: ``tan``.
            sdf_version : str
                The SDFormat version to use when constructing this element.

            Attributes
            ----------
            c1 : float
                See ``Parameters`` section.
            c2 : float
                See ``Parameters`` section.
            c3 : float
                See ``Parameters`` section.
            f : float
                See ``Parameters`` section.
            fun : str
                See ``Parameters`` section.

            """

            def __init__(
                self,
                *,
                c1: float = 1,
                c2: float = 1,
                c3: float = 0,
                f: float = 1,
                fun: str = "tan",
                sdf_version: str,
            ) -> None:
                super().__init__(sdf_version=sdf_version)
                self.c1 = (c1,)
                self.c2 = (c2,)
                self.c3 = (c3,)
                self.f = (f,)
                self.fun = fun

            @classmethod
            def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
                return Camera.Lense.CustomFunction(
                    c1=specific.c1,
                    c2=specific.c2,
                    c3=specific.c3,
                    f=specific.f,
                    fun=specific.fun,
                    sdf_version=version,
                )

        class Intrinsics(ElementBase):
            """Custom projection matrix.

            Camera intrinsic parameters for setting a custom perspective
            projection matrix (cannot be used with WideAngleCamera since this
            class uses image stitching from 6 different cameras for achieving a
            wide field of view). The focal lengths can be computed using
            focal_length_in_pixels = (image_width_in_pixels * 0.5) /
            tan(field_of_view_in_degrees * 0.5 * PI/180)

            Parameters
            ----------
            fx : float
                X focal length (in pixels, overrides horizontal_fov). Default: ``277``.
            fy : float
                Y focal length (in pixels, overrides horizontal_fov). Default: ``277``.
            cx : float
                X principal point (in pixels). Default: ``160``.
            cy : float
                Y principal point (in pixels). Default: ``120``.
            s : float
                XY axis skew. Default: ``0``.
            sdf_version : str
                The SDFormat version to use when constructing this element.

            Attributes
            ----------
            fx : float
                See ``Parameters`` section.
            fy : float
                See ``Parameters`` section.
            cx : float
                See ``Parameters`` section.
            cy : float
                See ``Parameters`` section.
            s : float
                See ``Parameters`` section.

            """

            def __init__(
                self,
                *,
                fx: float = 277,
                fy: float = 277,
                cx: float = 160,
                cy: float = 120,
                s: float = 0,
                sdf_version: str,
            ) -> None:
                super().__init__(sdf_version=sdf_version)
                self.fx = fx
                self.fy = fy
                self.cx = cx
                self.cy = cy
                self.s = s

            @classmethod
            def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
                return Camera.Intrinsics(
                    fx=specific.fx,
                    fy=specific.fy,
                    cx=specific.cx,
                    cy=specific.cy,
                    s=specific.s,
                    sdf_version=version,
                )



from dataclasses import dataclass, field
from typing import List, Optional
from .joint import Joint
from .link import Link

__NAMESPACE__ = "sdformat/v1.0/model.xsd"


@dataclass
class Model:
    class Meta:
        name = "model"

    origin: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    link: List[Link] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
            "min_occurs": 1,
        },
    )
    joint: List[Joint] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    plugin: List["Model.Plugin"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    gripper: List["Model.Gripper"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    static: bool = field(
        default=False,
        metadata={
            "type": "Attribute",
        },
    )

    @dataclass
    class Plugin:
        any_element: List[object] = field(
            default_factory=list,
            metadata={
                "type": "Wildcard",
                "namespace": "##any",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )
        filename: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

    @dataclass
    class Gripper:
        grasp_check: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        gripper_link: List[str] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
                "min_occurs": 1,
            },
        )
        palm_link: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )



import numpy as np
from typing import Any, Dict, Tuple
import warnings

from .... import transform as tf


def vector3(value: str):
    return np.fromstring(value, dtype=float, count=3, sep=" ")


def quaternion(value: str):
    # Note: SDFormat quaternions are XYZW
    return np.fromstring(value, dtype=float, count=4, sep=" ")


def vector2d(value: str):
    return np.fromstring(value, dtype=float, count=2, sep=" ")


def vector2i(value: str):
    return np.fromstring(value, dtype=int, count=2, sep=" ")


class ElementBase:
    def __init__(self, *, sdf_version: str) -> None:
        self.sdf_version = sdf_version

    @classmethod
    def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
        """Create from version-specific object

        This function converts a version-specific instantce of this element
        into its generic counterpart.

        Parameters
        ----------
        specific : Any
            The version-specific object to convert. It can come from any
            of scikit-bot's version-specific bindings.
        version : str
            The SDFormat version of the specific object.

        Returns
        -------
        generic : ElementBase
            The generic counterpart of the specific element.

        """
        warnings.warn(f"`{cls.__name__}` is not implemented yet.")
        return cls(sdf_version=version)

    def to_static_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        *,
        seed: int = None,
        shape: Tuple,
        axis: int = -1,
    ) -> tf.Frame:
        """Convert to transform graph

        Turns this element into a :mod:`skbot.transform` graph using the
        _static_ constraints defined by this element. The resulting graph places
        a :class:`tf.Frame <skbot.transform.Frame>` at the position of each
        joint, link, sensor, etc., defined by this Element. It then connects
        these frames using translation and rotation :class:`tf.Links
        <skbot.transform.Link>` in such a way that conversion between any two
        coordinate frames becomes possible.

        It differs from :func:`to_dynamic_graph` in that all links in this graph
        are rigid and that joint connections are not modelled explicitly.
        Further, the resulting graph will (if applicable) ignore any defined
        :class:`State` and instead model this element's initial configuration.
        As such, it can not be used for dynamic calculations, e.g., with
        :mod:`skbot.inverse_kinematics`.

        Parameters
        ----------
        declared_frames : Dict[str, tf.Frame]
            The frames declared in the scope in which this element has
            been defined.
        seed: bool
            The seed to use when procedually generating elements of the simulation.
            This is currently limited to the `Population` element only.
        shape : tuple
            A tuple describing the shape of elements that the resulting graph should
            transform. This can be used to add batch dimensions to the graph, for
            example to perform vectorized computation on multiple instances of the
            same world in different states, or for batched coordinate
            transformation. Defaults to (3,), which is the shape of a single 3D
            vector in euclidian space.
        axis : int
            The axis along which elements are stored. The axis must have length 3
            (since SDFormat describes 3 dimensional worlds), and all other axis are
            considered batch dimensions. Defaults to -1.

        Returns
        -------
        frame_graph : tf.Frame
            A frame graph modelling the (static) initial configuration of this
            element.

        Notes
        -----
        This function modifies ``declared_frames`` as a side-effect by appending
        any frames declared by this frame to it.

        """
        warnings.warn(
            f"`{self.__class__.__name__}` does not implement `to_static_graph` yet."
        )
        return tf.Frame(3, name="missing")

    def to_dynamic_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        *,
        seed: int = None,
        shape: Tuple,
        axis: int = -1,
        apply_state: bool = True,
        _scaffolding: Dict[str, tf.Frame],
    ) -> tf.Frame:
        """Convert to transform graph

        Turns this element into a :mod:`skbot.transform` graph using the _dynamic_
        constraints defined by this element. The resulting graph places a
        :class:`tf.Frame <skbot.transform.Frame>` at the position of each joint,
        link, sensor, etc., defined by this Element. It then connects these
        frames according to the dynamic constraints defined within this element,
        e.g., constraints from nested :class:`Joints <Joint>`.

        It differs from :func:`to_static_graph` in that all connections are
        dynamic. As a result, the resuling graph can have :class:`State` applied
        to it and it may be used for dynamic calculation, e.g., with
        :mod:`skbot.inverse_kinematics`.


        Parameters
        ----------
        declared_frames : Dict[str, tf.Frame]
            The frames declared in the scope in which this element has
            been defined.
        seed: bool
            The seed to use when procedually generating elements of the simulation.
            This is currently limited to the `Population` element only.
        shape : tuple
            A tuple describing the shape of elements that the resulting graph should
            transform. This can be used to add batch dimensions to the graph, for
            example to perform vectorized computation on multiple instances of the
            same world in different states, or for batched coordinate
            transformation. Defaults to (3,), which is the shape of a single 3D
            vector in euclidian space.
        axis : int
            The axis along which elements are stored. The axis must have length 3
            (since SDFormat describes 3 dimensional worlds), and all other axis are
            considered batch dimensions. Defaults to -1.
        apply_state: bool
            If True (default) each object's state is set according to `World.state`
            (if present).
        _scaffolding : Dict[str, tf.Frame]
            A dictionary of (connected) frames that used to determine the
            relative (initial) position of objects in the simulation. This
            parameter is for internal use and may be removed or change at any
            time without notice.


        Returns
        -------
        frame_graph : tf.Frame
            A frame graph modelling the dynamic structure of this element.

        Notes
        -----
        If this element contains children that are not dynamically connected to
        this element, they will be discarded.

        This function modifies ``declared_frames`` as a side-effect by appending
        any frames declared by this frame to it.

        """
        warnings.warn(
            f"`{self.__class__.__name__}` does not implement `to_dynamic_graph` yet."
        )
        return tf.Frame(3, name="missing")

    def declared_frames(self) -> Dict[str, tf.Frame]:
        """Frames contained in this element.

        Returns
        -------
        frame_dict : Dict[str, tf.Frame]
            A (namespaced) dictionary of frames declared within this element.
            The dict key of each frame is the full (namespaced) name of the
            element in the corresponding SDF.

        Notes
        -----
        Namespaces follow the SDFormat convention and are separated using `::`.

        """
        warnings.warn(
            f"`{self.__class__.__name__}` does not implement `declared_frames` yet."
        )
        return dict()

    @staticmethod
    def _prepare_standard_args(
        specific: Any,
        args_with_default: Dict[str, "ElementBase"] = None,
        list_args: Dict[str, Tuple[str, "ElementBase"]] = None,
        *,
        version: str,
    ) -> Dict[str, Any]:
        if args_with_default is None:
            args_with_default = dict()

        if list_args is None:
            list_args = dict()

        generic_args: Dict[str, Any] = dict()

        # convert arguments with default values
        for name, clazz in args_with_default.items():
            if not hasattr(specific, name):
                continue

            if clazz in [StringElement, BoolElement, FloatElement, IntegerElement]:
                value = getattr(specific, name)
                if value == "__default__":
                    continue
                elif value is not None:
                    generic_args[name] = value
                else:
                    continue
            elif getattr(specific, name) is None:
                generic_args[name] = clazz(sdf_version=version)
            else:
                generic_args[name] = clazz.from_specific(
                    getattr(specific, name), version=version
                )

        for their_name in list_args:
            our_name, clazz = list_args[their_name]

            if not hasattr(specific, their_name):
                generic_args[our_name] = []
                continue

            generic_args[our_name] = [
                clazz.from_specific(x, version=version)
                for x in getattr(specific, their_name)
            ]

        return generic_args


class StringElement(ElementBase):
    """Plumbing for smoother conversion"""


class FloatElement(ElementBase):
    """Plumbing for smoother conversion"""


class IntegerElement(ElementBase):
    """Plumbing for smoother conversion"""


class BoolElement(ElementBase):
    """Plumbing for smoother conversion"""


class Pose(ElementBase):
    """Position and orientation of a simulated object

    A position (x,y,z) and orientation (roll, pitch yaw) with respect
    to the frame named in the relative_to attribute.

    Parameters
    ----------
    value : str
        The numerical value of the pose. It is a string of 6 numbers
        separated by spaces. Default: "0 0 0 0 0 0"
    relative_to : str
        If specified, this pose is expressed in the named frame. The named frame
        must be declared within the same scope (world/model) as the element that
        has its pose specified by this tag.

        If missing, the pose is expressed in the frame of the parent XML element
        of the element that contains the pose. For exceptions to this rule and
        more details on the default behavior, see
        http://sdformat.org/tutorials?tut=pose_frame_semantics.

        Note that @relative_to merely affects an element's initial pose and
        does not affect the element's dynamic movement thereafter.

        .. versionadded:: SDFormat v1.8
            New in v1.8: @relative_to may use frames of nested scopes. In this
            case, the frame is specified using `::` as delimiter to define the
            scope of the frame, e.g.
            `nested_model_A::nested_model_B::awesome_frame`.
    frame : str
        Old name for `relative_to`.
        .. deprecated:: SDFormat v1.7


    """

    def __init__(
        self,
        *,
        value: str = "0 0 0 0 0 0",
        relative_to: str = None,
        frame: str = None,
        sdf_version: str,
    ) -> None:
        super().__init__(sdf_version=sdf_version)
        self.value = np.fromstring(value, dtype=float, count=6, sep=" ")
        self.relative_to = relative_to
        if frame is not None:
            self.relative_to = frame

        if self.relative_to == "":
            self.relative_to = None

    @property
    def frame(self):
        warnings.warn("`Pose.frame` is deprecated. Use `Pose.relative_to` instead.")
        return self.relative_to

    @classmethod
    def from_specific(cls, specific: Any, *, version: str):
        if specific is None:
            return Pose(sdf_version=version)
        elif version in ["1.0", "1.2", "1.3", "1.4"]:
            return Pose(value=specific, sdf_version=version)
        elif version in ["1.5", "1.6"]:
            return Pose(value=specific.value, frame=specific.frame, sdf_version=version)
        else:
            return Pose(
                value=specific.value,
                relative_to=specific.relative_to,
                sdf_version=version,
            )

    def to_tf_link(self) -> tf.Link:
        """tf.Link from **child** to **parent** frame"""
        offset = self.value[:3]
        angles = self.value[3:]

        return tf.CompundLink(
            [
                tf.EulerRotation("xyz", angles),
                tf.Translation(offset),
            ]
        )

    def to_static_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        child_name: str,
        *,
        shape: Tuple,
        axis: int = -1,
    ) -> None:
        parent_name = self.relative_to

        parent = declared_frames[parent_name]
        child = declared_frames[child_name]

        link = self.to_tf_link()
        link(child, parent)


class PoseBearing(ElementBase):
    def __init__(self, *, pose: Pose = None) -> None:
        self.pose = pose
        if self.pose is None:
            self.pose = Pose()


class NamedPoseBearing(PoseBearing):
    def __init__(self, *, name: str, pose: Pose = None) -> None:
        super().__init__(pose=pose)
        self.name = name


"""
Elements not yet represented:


<?xml version='1.0' encoding='UTF-8'?>
<xsd:schema xmlns:xsd='http://www.w3.org/2001/XMLSchema'>
  <xsd:simpleType name="time">
    <xsd:restriction base="xsd:string">
      <xsd:whiteSpace value="collapse"/>
      <xsd:pattern value="\d+ \d+"/>
    </xsd:restriction>
  </xsd:simpleType>

  <xsd:simpleType name="color">
    <xsd:restriction base="xsd:string">
      <xsd:pattern value="(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*"/>
    </xsd:restriction>
  </xsd:simpleType>
</xsd:schema>

"""



from dataclasses import dataclass, field
from typing import Optional
from .geometry import Geometry

__NAMESPACE__ = "sdformat/v1.0/visual.xsd"


@dataclass
class Visual:
    class Meta:
        name = "visual"

    origin: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    material: Optional["Visual.Material"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    geometry: Optional[Geometry] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    cast_shadows: bool = field(
        default=True,
        metadata={
            "type": "Attribute",
        },
    )
    laser_retro: float = field(
        default=0.0,
        metadata={
            "type": "Attribute",
        },
    )
    transparency: float = field(
        default=0.0,
        metadata={
            "type": "Attribute",
        },
    )

    @dataclass
    class Material:
        shader: Optional["Visual.Material.Shader"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        ambient: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        diffuse: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        specular: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        emissive: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        script: str = field(
            default="__default__",
            metadata={
                "type": "Attribute",
            },
        )

        @dataclass
        class Shader:
            normal_map: str = field(
                default="__default__",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            type: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )



"""
Interpolation of parameterized curves.

A collection of functions to compute positions along parameterized trajectories.
Here, trajectories are not limited to 3-dimensional space. Instead, they are
arbitrary mathematical paths with time parameterization. This means that a
trajectory may represent a sequence of positions in world space, but it may just 
as well represent a sequence of poses in joint space, or a sequence of
combined positions and velocities.

.. autosummary::
    :toctree:

    skbot.trajectory.linear_trajectory
    skbot.trajectory.spline_trajectory

"""

from .spline import spline_trajectory
from .linear import linear_trajectory
from . import utils

__all__ = ["spline_trajectory", "linear_trajectory", "utils"]



from dataclasses import dataclass, field
from typing import List, Optional

__NAMESPACE__ = "sdformat/v1.5/geometry.xsd"


@dataclass
class Geometry:
    """
    The shape of the visual or collision object.

    Parameters
    ----------
    empty: You can use the empty tag to make empty geometries.
    box: Box shape
    cylinder: Cylinder shape
    heightmap: A heightmap based on a 2d grayscale image.
    image: Extrude a set of boxes from a grayscale image.
    mesh: Mesh shape
    plane: Plane shape
    polyline: Defines an extruded polyline shape
    sphere: Sphere shape
    """

    class Meta:
        name = "geometry"

    empty: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    box: Optional["Geometry.Box"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    cylinder: Optional["Geometry.Cylinder"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    heightmap: Optional["Geometry.Heightmap"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    image: Optional["Geometry.Image"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    mesh: Optional["Geometry.Mesh"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    plane: Optional["Geometry.Plane"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    polyline: Optional["Geometry.Polyline"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    sphere: Optional["Geometry.Sphere"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )

    @dataclass
    class Box:
        """
        Box shape.

        Parameters
        ----------
        size: The three side lengths of the box. The origin of the box
            is in its geometric center (inside the center of the box).
        """

        size: str = field(
            default="1 1 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )

    @dataclass
    class Cylinder:
        """
        Cylinder shape.

        Parameters
        ----------
        radius: Radius of the cylinder
        length: Length of the cylinder along the z axis
        """

        radius: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        length: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Heightmap:
        """
        A heightmap based on a 2d grayscale image.

        Parameters
        ----------
        uri: URI to a grayscale image file
        size: The size of the heightmap in world units.       When
            loading an image: "size" is used if present, otherwise
            defaults to 1x1x1.       When loading a DEM: "size" is used
            if present, otherwise defaults to true size of DEM.
        pos: A position offset.
        texture: The heightmap can contain multiple textures. The order
            of the texture matters. The first texture will appear at the
            lowest height, and the last texture at the highest height.
            Use blend to control the height thresholds and fade between
            textures.
        blend: The blend tag controls how two adjacent textures are
            mixed. The number of blend elements should equal one less
            than the number of textures.
        use_terrain_paging: Set if the rendering engine will use terrain
            paging
        """

        uri: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        size: str = field(
            default="1 1 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        pos: str = field(
            default="0 0 0",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        texture: List["Geometry.Heightmap.Texture"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        blend: List["Geometry.Heightmap.Blend"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        use_terrain_paging: bool = field(
            default=False,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

        @dataclass
        class Texture:
            """The heightmap can contain multiple textures.

            The order of the texture matters. The first texture will
            appear at the lowest height, and the last texture at the
            highest height. Use blend to control the height thresholds
            and fade between textures.

            Parameters
            ----------
            size: Size of the applied texture in meters.
            diffuse: Diffuse texture image filename
            normal: Normalmap texture image filename
            """

            size: float = field(
                default=10.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            diffuse: str = field(
                default="__default__",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            normal: str = field(
                default="__default__",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

        @dataclass
        class Blend:
            """The blend tag controls how two adjacent textures are mixed.

            The number of blend elements should equal one less than the
            number of textures.

            Parameters
            ----------
            min_height: Min height of a blend layer
            fade_dist: Distance over which the blend occurs
            """

            min_height: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            fade_dist: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

    @dataclass
    class Image:
        """
        Extrude a set of boxes from a grayscale image.

        Parameters
        ----------
        uri: URI of the grayscale image file
        scale: Scaling factor applied to the image
        threshold: Grayscale threshold
        height: Height of the extruded boxes
        granularity: The amount of error in the model
        """

        uri: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        scale: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        threshold: int = field(
            default=200,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        height: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        granularity: int = field(
            default=1,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Mesh:
        """
        Mesh shape.

        Parameters
        ----------
        uri: Mesh uri
        submesh: Use a named submesh. The submesh must exist in the mesh
            specified by the uri
        scale: Scaling factor applied to the mesh
        """

        uri: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        submesh: Optional["Geometry.Mesh.Submesh"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        scale: str = field(
            default="1 1 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )

        @dataclass
        class Submesh:
            """Use a named submesh.

            The submesh must exist in the mesh specified by the uri

            Parameters
            ----------
            name: Name of the submesh within the parent mesh
            center: Set to true to center the vertices of the submesh at
                0,0,0. This will effectively remove any transformations
                on the submesh before the poses from parent links and
                models are applied.
            """

            name: str = field(
                default="__default__",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            center: bool = field(
                default=False,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

    @dataclass
    class Plane:
        """
        Plane shape.

        Parameters
        ----------
        normal: Normal direction for the plane
        size: Length of each side of the plane
        """

        normal: str = field(
            default="0 0 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        size: str = field(
            default="1 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+)((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )

    @dataclass
    class Polyline:
        """
        Defines an extruded polyline shape.

        Parameters
        ----------
        point: A series of points that define the path of the polyline.
        height: Height of the polyline
        """

        point: List[str] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
                "min_occurs": 1,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+)((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        height: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Sphere:
        """
        Sphere shape.

        Parameters
        ----------
        radius: radius of the sphere
        """

        radius: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
